import abc

from gen_epix.fastapp import BaseRepository
from gen_epix.fastapp.unit_of_work import BaseUnitOfWork
from gen_epix.omopdb.domain import model  # forces models to be registered now
from gen_epix.omopdb.domain import DOMAIN
from gen_epix.omopdb.domain.enum import ServiceType


class BaseOrganizationRepository(BaseRepository):
    ENTITIES = DOMAIN.get_dag_sorted_entities(
        service_type=ServiceType.ORGANIZATION, persistable=True
    )

    @abc.abstractmethod
    def retrieve_user_by_key(self, uow: BaseUnitOfWork, user_key: str) -> model.User:
        raise NotImplementedError()



# pylint: disable=useless-import-alias
from gen_epix.seqdb.repositories.organization_dict import (
    OrganizationDictRepository as OrganizationDictRepository,
)
from gen_epix.seqdb.repositories.organization_sa import (
    OrganizationSARepository as OrganizationSARepository,
)
from gen_epix.seqdb.repositories.seq_dict import SeqDictRepository as SeqDictRepository
from gen_epix.seqdb.repositories.seq_sa import SeqSARepository as SeqSARepository
from gen_epix.seqdb.repositories.system_dict import (
    SystemDictRepository as SystemDictRepository,
)
from gen_epix.seqdb.repositories.system_sa import (
    SystemSARepository as SystemSARepository,
)



import datetime
from decimal import Decimal
from typing import Any, Callable, Iterable, Type
from uuid import UUID

import gen_epix.casedb.domain.command as command
import gen_epix.casedb.domain.enum as enum
import gen_epix.casedb.domain.model as model
from gen_epix.casedb.domain import exc
from gen_epix.casedb.domain.policy.abac import BaseCaseAbacPolicy
from gen_epix.casedb.domain.service import BaseCaseService
from gen_epix.fastapp import BaseUnitOfWork, CrudOperation
from gen_epix.fastapp.enum import CrudOperationSet
from gen_epix.filter import (
    BooleanOperator,
    CompositeFilter,
    DatetimeRangeFilter,
    Filter,
    StringSetFilter,
    UuidSetFilter,
)
from util.util import map_paired_elements


class CaseService(BaseCaseService):
    _VALUE_TO_STR = {
        enum.ColType.TIME_DAY: lambda x: None if not x else f"{x}",
        enum.ColType.TIME_WEEK: lambda x: None if not x else f"{x}",
        enum.ColType.TIME_MONTH: lambda x: None if not x else f"{x}",
        enum.ColType.TIME_QUARTER: lambda x: None if not x else f"{x}",
        enum.ColType.TIME_YEAR: lambda x: None if not x else f"{x}",
        enum.ColType.GEO_LATLON: lambda x: None if not x else f"{x}",
        enum.ColType.TEXT: lambda x: None if not x else f"{x}",
        enum.ColType.ID_DIRECT: lambda x: None if not x else f"{x}",
        enum.ColType.ID_PSEUDONYMISED: lambda x: None if not x else f"{x}",
        enum.ColType.OTHER: lambda x: None if not x else f"{x}",
        enum.ColType.DECIMAL_0: lambda x: (
            None if not x else (x if isinstance(x, str) else f"{x:.0f}")
        ),
        enum.ColType.DECIMAL_1: lambda x: (
            None if not x else (x if isinstance(x, str) else f"{x:.1f}")
        ),
        enum.ColType.DECIMAL_2: lambda x: (
            None if not x else (x if isinstance(x, str) else f"{x:.2f}")
        ),
        enum.ColType.DECIMAL_3: lambda x: (
            None if not x else (x if isinstance(x, str) else f"{x:.3f}")
        ),
        enum.ColType.DECIMAL_4: lambda x: (
            None if not x else (x if isinstance(x, str) else f"{x:.4f}")
        ),
        enum.ColType.DECIMAL_5: lambda x: (
            None if not x else (x if isinstance(x, str) else f"{x:.5f}")
        ),
        enum.ColType.DECIMAL_6: lambda x: (
            None if not x else (x if isinstance(x, str) else f"{x:.6f}")
        ),
    }

    def crud(  # type:ignore[override]
        self, cmd: command.CrudCommand
    ) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
        """
        Override the base crud method to apply ABAC restrictions and cascade delete
        where necessary
        """
        # Handle no ABAC restrictions
        if any(isinstance(cmd, x) for x in BaseCaseService.NO_ABAC_COMMAND_CLASSES):
            # No ABAC restrictions
            return super().crud(cmd)  # type: ignore[return-value]

        # Start unit of work and execute all within this scope
        with self.repository.uow() as uow:
            # Metadata commands
            if any(
                isinstance(cmd, x)
                for x in BaseCaseService.ABAC_METADATA_COMMAND_CLASSES
            ):
                return self._crud_metadata(uow, cmd)  # type: ignore[no-any-return]
            # Data commands
            elif any(
                isinstance(cmd, x) for x in BaseCaseService.ABAC_DATA_COMMAND_CLASSES
            ):
                return self._crud_data(uow, cmd)
            else:
                raise AssertionError(
                    f"Unexpected command {cmd.__class__.__name__} with operation {cmd.operation.value}"
                )

    def create_cases_or_set(
        self, cmd: command.CaseSetCreateCommand | command.CasesCreateCommand
    ) -> model.CaseSet | list[model.Case] | None:
        if isinstance(cmd, command.CaseSetCreateCommand):
            is_case_set = True
        elif isinstance(cmd, command.CasesCreateCommand):
            is_case_set = False
        else:
            raise AssertionError(
                f"Unexpected command {cmd.__class__.__name__} with operation {cmd.operation.value}"
            )

        # Special case: zero cases to be created
        if not is_case_set and len(cmd.cases) == 0:  # type: ignore[union-attr]
            return []

        # Get case type and created_in data collection IDs
        if is_case_set:
            case_type_id = cmd.case_set.case_type_id  # type: ignore[union-attr]
            created_in_data_collection_id = cmd.case_set.created_in_data_collection_id  # type: ignore[union-attr]
        else:
            case_type_id = cmd.cases[0].case_type_id  # type: ignore[union-attr]
            created_in_data_collection_id = cmd.cases[0].created_in_data_collection_id  # type: ignore[union-attr]

        # @ABAC: verify if case set or cases may be created in the given data collection(s):
        # exactly one private data collection must be provided, and add_case must be true
        # for all the other data collections
        # TODO: add checking for equivalent share rights establishing a connected path to these data collections
        case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
        assert case_abac is not None
        is_allowed = case_abac.is_allowed(
            case_type_id,
            (enum.CaseRight.ADD_CASE_SET if is_case_set else enum.CaseRight.ADD_CASE),
            True,
            created_in_data_collection_id=created_in_data_collection_id,
            tgt_data_collection_ids=cmd.data_collection_ids,
        )
        if not is_allowed:
            assert cmd.user is not None
            raise exc.UnauthorizedAuthError(
                f"User {cmd.user.id} is not allowed to create a case set/cases in the given data collection(s)"
            )

        # Create case set/cases, case set/case data collection links, and optionally
        # case set members
        with self.repository.uow() as uow:
            # Create case set/cases, using the parent class method to avoid ABAC
            # restrictions
            if is_case_set:
                case_set: model.CaseSet = super().crud(  # type: ignore[assignment]
                    command.CaseSetCrudCommand(
                        user=cmd.user,
                        operation=CrudOperation.CREATE_ONE,
                        objs=cmd.case_set,  # type: ignore[union-attr,arg-type]
                    )
                )
            else:
                cases: list[model.Case] = super().crud(  # type: ignore[assignment]
                    command.CaseCrudCommand(
                        user=cmd.user,
                        operation=CrudOperation.CREATE_SOME,
                        objs=cmd.cases,  # type: ignore[union-attr,arg-type]
                    )
                )
            # Associate case set/cases with data collections
            if is_case_set:
                assert case_set.id is not None
                curr_cmd: command.CaseSetDataCollectionLinkCrudCommand = (
                    command.CaseSetDataCollectionLinkCrudCommand(
                        user=cmd.user,
                        operation=CrudOperation.CREATE_SOME,
                        objs=[
                            model.CaseSetDataCollectionLink(
                                case_set_id=case_set.id, data_collection_id=x
                            )
                            for x in cmd.data_collection_ids
                        ],
                    )
                )
            else:
                curr_cmd: command.CaseDataCollectionLinkCrudCommand = (  # type: ignore[no-redef]
                    command.CaseDataCollectionLinkCrudCommand(
                        user=cmd.user,
                        operation=CrudOperation.CREATE_SOME,
                        objs=[
                            model.CaseDataCollectionLink(
                                case_id=x.id, data_collection_id=y  # type: ignore[arg-type]
                            )
                            for x in cases
                            for y in cmd.data_collection_ids
                        ],
                    )
                )
            curr_cmd._policies.extend(cmd._policies)
            cases_or_set_data_collection_links = self.crud(curr_cmd)
            # Associate case set with cases if necessary
            if is_case_set and cmd.case_ids:  # type: ignore[union-attr]
                curr_cmd2: command.CaseSetMemberCrudCommand = (
                    command.CaseSetMemberCrudCommand(
                        user=cmd.user,
                        operation=CrudOperation.CREATE_SOME,
                        objs=[
                            model.CaseSetMember(case_set_id=case_set.id, case_id=x)  # type: ignore[arg-type]
                            for x in cmd.case_ids  # type: ignore[union-attr]
                        ],
                        _policies=cmd._policies,
                    )
                )
                curr_cmd2._policies.extend(cmd._policies)
                case_set_members = self.crud(curr_cmd2)
        return case_set if is_case_set else cases

    def retrieve_complete_case_type(
        self,
        cmd: command.RetrieveCompleteCaseTypeCommand,
    ) -> model.CompleteCaseType:
        # TODO: many calls are inefficient,
        # retrieving first all objs and then filtering.
        # To be improved with e.g. CQS.
        user, repository = self._get_user_and_repository(cmd)

        with repository.uow() as uow:
            # Get case type
            case_type_id = cmd.case_type_id
            case_type: model.CaseType = self.repository.crud(  # type: ignore[assignment]
                uow,
                user.id,
                model.CaseType,
                None,
                case_type_id,
                CrudOperation.READ_ONE,
            )

            # @ABAC
            # Get allowed case type columns with any CRUD permission
            case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
            assert case_abac is not None
            case_type_access_abacs: dict[UUID, model.CaseTypeAccessAbac] = (
                case_abac.case_type_access_abacs.get(case_type_id, {})
            )
            case_type_share_abacs: dict[UUID, model.CaseTypeShareAbac] = (
                case_abac.case_type_share_abacs.get(case_type_id, {})
            )

            abac_case_type_col_ids: set[UUID]
            if case_abac.is_full_access:
                # Special case: full access -> all rights for all data collections for
                # this case type
                # TODO: consider if it should be limited to the union of all the
                # organization rights instead. A root user e.g. may then still have
                # full access by using the CRUD methods
                abac_case_type_col_ids = repository.crud(  # type: ignore[assignment]
                    uow,
                    user.id,
                    model.CaseTypeCol,
                    None,
                    None,
                    CrudOperation.READ_ALL,
                    filter=UuidSetFilter(
                        key="case_type_id",
                        members=frozenset({case_type_id}),
                    ),
                    return_id=True,
                )
                data_collection_ids: list[UUID] = self.app.handle(
                    command.DataCollectionCrudCommand(
                        user=user,
                        operation=CrudOperation.READ_ALL,
                        props={"return_id": True},
                    )
                )
                case_type_access_abacs = {
                    x: model.CaseTypeAccessAbac(
                        case_type_id=case_type_id,
                        data_collection_id=x,
                        is_private=True,
                        add_case=True,
                        remove_case=True,
                        read_case_type_col_ids=abac_case_type_col_ids,
                        write_case_type_col_ids=abac_case_type_col_ids,
                        add_case_set=True,
                        remove_case_set=True,
                        read_case_set=True,
                        write_case_set=True,
                    )
                    for x in data_collection_ids
                }
                # case_type_share_abacs can be empty since all rights are already in
                # case_type_access_abacs
                case_type_share_abacs = {}
            else:
                abac_case_type_col_ids = set()
                for x in case_type_access_abacs.values():
                    abac_case_type_col_ids.update(x.read_case_type_col_ids)
                    abac_case_type_col_ids.update(x.write_case_type_col_ids)

            # Get etiologies
            if case_type.disease_id:
                etiologies = self.app.handle(
                    command.EtiologyCrudCommand(
                        user=user,
                        operation=CrudOperation.READ_ALL,
                    )
                )
                etiologies = {
                    x.id: x for x in etiologies if x.disease_id == case_type.disease_id
                }
            else:
                etiologies = {}

            # Get etiological agents
            if etiologies:
                etiological_agent_ids = list(
                    x.etiological_agent_id for x in etiologies.values()
                )
                etiological_agents = self.app.handle(
                    command.EtiologicalAgentCrudCommand(
                        user=user,
                        operation=CrudOperation.READ_SOME,
                        obj_ids=etiological_agent_ids,
                    )
                )
                etiological_agents = {x.id: x for x in etiological_agents}
            else:
                etiological_agents = {}

            # Get allowed case_type_cols
            case_type_col_ids = list(abac_case_type_col_ids)
            case_type_cols_: list[model.CaseTypeCol] = repository.crud(  # type: ignore[assignment]
                uow,
                user.id,
                model.CaseTypeCol,
                None,
                case_type_col_ids,
                CrudOperation.READ_SOME,
            )
            case_type_cols: dict[UUID, model.CaseTypeCol] = {
                x.id: x for x in case_type_cols_  # type: ignore[misc]
            }

            # # Special case: no case_type_cols
            # if not case_type_cols:
            #     return model.CompleteCaseType(
            #         **case_type.model_dump(),
            #         etiologies=etiologies,
            #         etiological_agents=etiological_agents,
            #         dims={},
            #         cols={},
            #         case_type_dims=[],
            #         case_type_cols={},
            #         case_type_col_order=[],
            #         genetic_distance_protocols={},
            #         tree_algorithms={},
            #         case_type_access_abacs=case_type_access_abacs,
            #     )

            # Get cols
            col_ids = list({x.col_id for x in case_type_cols.values()})
            cols_: list[model.Col] = repository.crud(  # type: ignore[assignment]
                uow,
                user.id,
                model.Col,
                None,
                col_ids,
                CrudOperation.READ_SOME,
            )
            cols: dict[UUID, model.Col] = {x.id: x for x in cols_}  # type: ignore[misc]

            # Get dims
            dim_ids = list({x.dim_id for x in cols.values()})
            dims_: list[model.Dim] = repository.crud(  # type: ignore[assignment]
                uow,
                user.id,
                model.Dim,
                None,
                dim_ids,
                CrudOperation.READ_SOME,
            )
            dims: dict[UUID, model.Dim] = {x.id: x for x in dims_}  # type: ignore[misc]

            # Get case_type_col_order
            # TODO: to be tested
            max_dim_rank = max([0] + [x.rank for x in dims.values() if x.rank])
            max_col_rank_in_dim = max(
                [0] + [x.rank_in_dim for x in cols.values() if x.rank_in_dim]
            )
            max_case_type_col_rank = max(
                [0] + [x.rank for x in case_type_cols.values() if x.rank]
            )
            max_case_type_col_occurrence = max(
                [0] + [x.occurrence for x in case_type_cols.values() if x.occurrence]
            )
            case_type_col_keys: dict[UUID, tuple[int, int, int]] = {
                x.id: (  # type: ignore[misc]
                    x.rank if x.rank else max_case_type_col_rank,
                    (
                        dims[cols[x.col_id].dim_id].rank
                        if dims[cols[x.col_id].dim_id].rank
                        else max_dim_rank
                    ),
                    (
                        cols[x.col_id].rank_in_dim
                        if cols[x.col_id].rank_in_dim
                        else max_col_rank_in_dim
                    ),
                    x.occurrence if x.occurrence else max_case_type_col_occurrence,
                )
                for x in case_type_cols.values()
            }
            case_type_col_order = list(case_type_col_keys.keys())
            case_type_col_order.sort(key=lambda x: case_type_col_keys[x])

            # Get case_type_dims as the list ordered by the (dim, occurrence)
            # that occurs first in case_type_col_order
            dict_: dict[tuple[UUID, int | None], list] = {}
            # dict[tuple[dim_id, occurrence], list[rank, [tuple[case_type_col_id, col.rank_in_dim]]]]
            for case_type_col_id in case_type_col_order:
                # Add to dict_
                case_type_col = case_type_cols[case_type_col_id]
                col = cols[case_type_col.col_id]
                tuple_ = (col.dim_id, case_type_col.occurrence)
                if tuple_ in dict_:
                    dict_[tuple_][1].append((case_type_col_id, col.rank_in_dim))
                    continue
                dict_[tuple_] = [len(dict_), [(case_type_col_id, col.rank_in_dim)]]
            case_type_dim_order = list(dict_.keys())
            case_type_dim_order.sort(key=lambda x: dict_[x][0])
            case_type_dims = [
                model.CaseTypeDim(
                    id=x[0],
                    dim_id=x[0],
                    occurrence=x[1],
                    rank=i + 1,
                    case_type_col_order=[],
                )
                for i, x in enumerate(case_type_dim_order)
            ]
            for case_type_dim in case_type_dims:
                # Fill in id and case_type_col_order
                tuples = dict_[(case_type_dim.dim_id, case_type_dim.occurrence)][1]
                tuples.sort(key=lambda x: 1 if x[1] is None else x[1])
                case_type_dim.case_type_col_order = [x[0] for x in tuples]
                case_type_dim.id = case_type_dim.case_type_col_order[0]

            # Get genetic distance protocols
            genetic_distance_protocols = self.app.handle(
                command.GeneticDistanceProtocolCrudCommand(
                    user=user,
                    operation=CrudOperation.READ_SOME,
                    obj_ids=list(
                        {
                            x.genetic_distance_protocol_id
                            for x in cols.values()
                            if x.genetic_distance_protocol_id
                        }
                    ),
                )
            )
            genetic_distance_protocols = {x.id: x for x in genetic_distance_protocols}

            # Get tree algorithms
            tree_algorithm_codes = set.union(
                set(),
                *[
                    x.tree_algorithm_codes
                    for x in case_type_cols.values()
                    if x.tree_algorithm_codes
                ],
            )
            tree_algorithms = self.app.handle(
                command.TreeAlgorithmCrudCommand(
                    user=user,
                    operation=CrudOperation.READ_ALL,
                )
            )
            tree_algorithms = {
                x.code: x for x in tree_algorithms if x.code in tree_algorithm_codes
            }

        # Compose complete case type and return
        return model.CompleteCaseType(
            **case_type.model_dump(),
            etiologies=etiologies,
            etiological_agents=etiological_agents,
            dims=dims,
            cols=cols,
            case_type_dims=case_type_dims,
            case_type_cols=case_type_cols,
            case_type_col_order=case_type_col_order,
            genetic_distance_protocols=genetic_distance_protocols,
            tree_algorithms=tree_algorithms,
            case_type_access_abacs=case_type_access_abacs,
            case_type_share_abacs=case_type_share_abacs,
        )

    def retrieve_case_type_stats(
        self,
        cmd: command.RetrieveCaseTypeStatsCommand,
    ) -> list[model.CaseTypeStat]:
        user, repository = self._get_user_and_repository(cmd)
        case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
        assert case_abac is not None
        case_type_ids = cmd.case_type_ids
        with repository.uow() as uow:
            cases: list[model.Case] = self._retrieve_cases_with_content_right(
                uow,
                user.id,
                case_abac,
                # user_case_access,
                enum.CaseRight.READ_CASE,
                datetime_range_filter=cmd.datetime_range_filter,
                filter_content=False,
            )
            if case_type_ids is not None:
                cases = [x for x in cases if x.case_type_id in case_type_ids]
            else:
                case_type_ids = {x.case_type_id for x in cases}
            # Derive stats
            empty_stat = {
                "n_cases": 0,
                "first_case_month": None,
                "last_case_month": None,
            }
            stats = {x: dict(empty_stat) for x in case_type_ids}
            for case in cases:
                case_type_id = case.case_type_id
                date_ = case.case_date
                stat = stats[case_type_id]
                if stat["n_cases"] == 0:
                    stat["n_cases"] = 1
                    stat["first_case_month"] = date_  # type: ignore[assignment]
                    stat["last_case_month"] = date_  # type: ignore[assignment]
                else:
                    stat["n_cases"] += 1
                    stat["first_case_month"] = min(stat["first_case_month"], date_)  # type: ignore[type-var,assignment]
                    stat["last_case_month"] = max(stat["last_case_month"], date_)  # type: ignore[type-var,assignment]
            # Convert first/last date to month only
            for stat in stats.values():
                for key in ("first_case_month", "last_case_month"):
                    stat[key] = stat[key].isoformat()[0:7]  # type: ignore[union-attr]
            # Get case type stats
            case_type_stats = [
                model.CaseTypeStat(case_type_id=x, **stats[x]) for x in case_type_ids  # type: ignore[arg-type]
            ]
        return case_type_stats

    def retrieve_case_set_stats(
        self,
        cmd: command.RetrieveCaseSetStatsCommand,
    ) -> list[model.CaseSetStat]:
        user, repository = self._get_user_and_repository(cmd)
        case_set_ids = cmd.case_set_ids
        # Create filter, even if no case_set_ids are provided, to avoid unallowed read
        # all without filter
        query_filter: Filter | None = None
        if case_set_ids:
            query_filter = UuidSetFilter(
                key="case_set_id", members=cmd.case_set_ids  # type: ignore[arg-type]
            )
        with self.repository.uow() as uow:
            curr_cmd = command.CaseSetMemberCrudCommand(
                user=user,
                operation=CrudOperation.READ_ALL,
                query_filter=query_filter,
            )
            curr_cmd._policies.extend(cmd._policies)
            case_set_members: list[model.CaseSetMember] = self.crud(curr_cmd)  # type: ignore[assignment]
            case_set_case_ids: dict[UUID, set[UUID]] = map_paired_elements(  # type: ignore[assignment]
                ((x.case_set_id, x.case_id) for x in case_set_members), as_set=True
            )
            if not case_set_ids:
                case_set_ids = list(case_set_case_ids.keys())
            # Get cases
            # @ABAC: case_set_case_ids is already filtered on cases with access, no
            # need to apply here again
            cases_: list[model.Case] = self.repository.crud(  # type: ignore[assignment]
                uow,
                user.id,
                model.Case,
                None,
                list(set.union(set(), *list(case_set_case_ids.values()))),
                CrudOperation.READ_SOME,
            )
            cases = {x.id: x for x in cases_}
            # Create case set stats
            case_set_stats = []
            case_dates = {x.id: x.case_date for x in cases.values()}
            all_case_ids = set(cases.keys())
            for case_set_id in case_set_ids:
                case_ids = case_set_case_ids.get(case_set_id, set()).intersection(
                    all_case_ids
                )
                # TODO: calculate n_own_cases as the number of cases with a created_in data collection that is associated with the user
                n_own_cases = 0
                first_case_month = (
                    min(case_dates[x] for x in case_ids).isoformat()[0:7]
                    if case_ids
                    else None
                )
                last_case_month = (
                    max(case_dates[x] for x in case_ids).isoformat()[0:7]
                    if case_ids
                    else None
                )
                case_set_stats.append(
                    model.CaseSetStat(
                        case_set_id=case_set_id,
                        n_cases=len(case_ids),
                        n_own_cases=n_own_cases,
                        first_case_month=first_case_month,
                        last_case_month=last_case_month,
                    )
                )

        return case_set_stats

    def retrieve_cases_by_query(
        self, cmd: command.RetrieveCasesByQueryCommand
    ) -> list[UUID]:
        # TODO: This is an inefficient call first loading all cases, then filtering them and then keeping only the ids. To be replaced by optimized query.
        user, repository = self._get_user_and_repository(cmd)
        case_query = cmd.case_query
        case_set_ids = case_query.case_set_ids
        case_type_ids = case_query.case_type_ids
        datetime_range_filter = case_query.datetime_range_filter

        # Special case: zero case_set_ids or zero case_type_ids (None equals all)
        if case_set_ids is not None and len(case_set_ids) == 0:
            return []
        if case_type_ids is not None and len(case_type_ids) == 0:
            return []

        # @ABAC: get case abac
        case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
        assert case_abac is not None
        is_full_access = case_abac.is_full_access
        has_case_read = case_abac.get_combinations_with_access_right(
            enum.CaseRight.READ_CASE
        )

        # @ABAC: Verify read access to all given case types if applicable
        if case_type_ids and not is_full_access:
            if not case_type_ids.issubset(set(has_case_read.keys())):
                raise exc.UnauthorizedAuthError(
                    f"Unauthorized case types: {case_type_ids}"
                )

        with repository.uow() as uow:

            # @ABAC: Verify any access to all given case sets if applicable
            if case_set_ids:
                case_sets = self._retrieve_case_sets_with_content_right(
                    uow,
                    user.id,
                    case_abac,
                    # user_case_access
                    enum.CaseRight.READ_CASE_SET,
                ) + self._retrieve_case_sets_with_content_right(
                    uow,
                    user.id,
                    case_abac,
                    # user_case_access
                    enum.CaseRight.WRITE_CASE_SET,
                )
                invalid_case_set_ids = case_set_ids - {x.id for x in case_sets}
                if invalid_case_set_ids:
                    invalid_case_set_ids_str = ", ".join(
                        [str(x) for x in invalid_case_set_ids]
                    )
                    raise exc.UnauthorizedAuthError(
                        f"Unauthorized case sets: {invalid_case_set_ids_str}"
                    )

            # @ABAC: Verify validity of filter
            if case_query.filter:
                # Make sure filter keys are UUIDs
                case_query.filter.set_keys(
                    lambda x: UUID(x) if isinstance(x, str) else x
                )
                cols = self._verify_case_filter(uow, user, case_query.filter)

            # @ABAC: Retrieve all cases with read access, and content filtered on case type
            # col read access
            cases = self._retrieve_cases_with_content_right(
                uow,
                user.id,
                case_abac,
                # user_case_access,
                enum.CaseRight.READ_CASE,
                case_ids=None,
                datetime_range_filter=datetime_range_filter,
                filter_content=True,
            )

            # Filter cases by case types
            if case_type_ids:
                cases = [x for x in cases if x.case_type_id in case_type_ids]

            # Filter cases by case sets
            if case_set_ids:
                case_case_sets, _ = self._retrieve_case_case_sets_map(uow, user.id)
                cases = [
                    x
                    for x in cases
                    if x.id in case_case_sets
                    and case_case_sets[x.id].intersection(case_set_ids)
                ]

            # Filter cases by filters
            if case_query.filter:
                map_funs = CaseService._get_map_funs_for_filters(cols)
                cases = [
                    x
                    for x, y in zip(
                        cases,
                        case_query.filter.match_rows(
                            (x.content for x in cases), map_fun=map_funs  # type: ignore[misc]
                        ),
                    )
                    if y
                ]

        # TODO: consider putting these cases, with their data already filtered, in a
        # cache, so that the expected subsequent call to retrieve them can be sped up

        # Return case ids
        case_ids = [x.id for x in cases]
        return case_ids  # type: ignore[return-value]

    def retrieve_cases_by_id(
        self, cmd: command.RetrieveCasesByIdCommand
    ) -> list[model.Case]:
        case_ids = cmd.case_ids
        user, repository = self._get_user_and_repository(cmd)
        if not case_ids:
            return []
        # @ABAC: get case abac
        case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
        assert case_abac is not None

        with repository.uow() as uow:
            cases = self._retrieve_cases_with_content_right(
                uow,
                user.id,
                case_abac,
                enum.CaseRight.READ_CASE,
                case_ids=case_ids,
                filter_content=True,
            )
        return cases

    def retrieve_case_or_set_rights(
        self,
        cmd: command.RetrieveCaseRightsCommand | command.RetrieveCaseSetRightsCommand,
    ) -> list[model.CaseRights] | list[model.CaseSetRights]:
        is_case_set = isinstance(cmd, command.RetrieveCaseSetRightsCommand)
        case_or_set_ids = cmd.case_set_ids if is_case_set else cmd.case_ids  # type: ignore[union-attr]
        user, repository = self._get_user_and_repository(cmd)

        # Special case: zero case_ids
        if not case_or_set_ids:
            return []

        # @ABAC: get case abac
        case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
        assert case_abac is not None

        # Retrieve all cases and case data collection links
        with repository.uow() as uow:
            # Retrieve case/set data collection links
            crud_command_class = (
                command.CaseSetDataCollectionLinkCrudCommand
                if is_case_set
                else command.CaseDataCollectionLinkCrudCommand
            )
            key = "case_set_id" if is_case_set else "case_id"
            curr_cmd = crud_command_class(
                user=user,
                operation=CrudOperation.READ_ALL,
                query_filter=UuidSetFilter(key=key, members=case_or_set_ids),  # type: ignore[arg-type]
            )
            curr_cmd._policies.extend(cmd._policies)
            case_or_set_data_collection_links: list[model.CaseDataCollectionLink] | list[model.CaseSetDataCollectionLink] = self.crud(curr_cmd)  # type: ignore[assignment]
            # Check validitity of case ids
            valid_case_or_set_ids = (
                {x.case_set_id for x in case_or_set_data_collection_links}  # type: ignore[union-attr]
                if is_case_set
                else {x.case_id for x in case_or_set_data_collection_links}  # type: ignore[union-attr]
            )
            invalid_case_or_set_ids = set(case_or_set_ids) - valid_case_or_set_ids
            if invalid_case_or_set_ids:
                raise exc.UnauthorizedAuthError(
                    f"Unauthorized case{' set' if is_case_set else ''} ids: {invalid_case_or_set_ids}"
                )
            # Retrieve cases/sets
            cases_or_sets: list[model.CaseSet] | list[model.Case] = self.repository.crud(  # type: ignore[assignment]
                uow,
                user.id,
                model.CaseSet if is_case_set else model.Case,
                None,
                case_or_set_ids,
                CrudOperation.READ_SOME,
            )

        # Determine case/set rights
        case_or_set_data_collections: dict[UUID, frozenset[UUID]] = map_paired_elements(  # type: ignore[assignment]
            (
                (x.case_set_id if is_case_set else x.case_id, x.data_collection_id)  # type: ignore[union-attr]
                for x in case_or_set_data_collection_links
            ),
            as_set=True,
            frozen=True,
        )

        # Generate return value
        retval: list[model.CaseSetRights] | list[model.CaseRights] = []
        for case_or_set in cases_or_sets:
            assert case_or_set.id is not None
            args: tuple = (
                case_or_set.id,
                case_or_set.case_type_id,
                case_or_set.created_in_data_collection_id,
                case_or_set_data_collections.get(case_or_set.id, set()),
            )
            retval.append(case_abac.get_case_set_rights(*args) if is_case_set else case_abac.get_case_rights(*args))  # type: ignore[arg-type]

        return retval

    def retrieve_phylogenetic_tree(
        self, cmd: command.RetrievePhylogeneticTreeByCasesCommand
    ) -> model.PhylogeneticTree:
        dist_case_type_col_id = cmd.genetic_distance_case_type_col_id
        tree_algorithm_code = cmd.tree_algorithm
        case_ids = cmd.case_ids
        user: model.User
        user, repository = self._get_user_and_repository(cmd)  # type: ignore[assignment]
        assert user.id is not None
        case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
        assert case_abac is not None

        with repository.uow() as uow:
            # Get distance column data
            dist_case_type_col: model.CaseTypeCol = repository.crud(  # type: ignore[assignment]
                uow,
                user.id,
                model.CaseTypeCol,
                None,
                dist_case_type_col_id,
                CrudOperation.READ_ONE,
            )
            case_type_id = dist_case_type_col.case_type_id
            dist_col: model.Col = repository.crud(  # type: ignore[assignment]
                uow,
                user.id,
                model.Col,
                None,
                dist_case_type_col.col_id,
                CrudOperation.READ_ONE,
            )
            if dist_col.col_type != enum.ColType.GENETIC_DISTANCE:
                raise exc.InvalidArgumentsError(
                    f"Case type column {dist_case_type_col_id} is not of type {enum.ColType.GENETIC_DISTANCE.value}"
                )
            # Get sequence column data
            seq_case_type_col_id = dist_case_type_col.genetic_sequence_case_type_col_id
            if not seq_case_type_col_id:
                raise exc.InvalidArgumentsError(
                    f"Case type column {dist_case_type_col_id} has no associated sequence column"
                )

            # @ABAC
            assert dist_case_type_col.tree_algorithm_codes is not None
            if tree_algorithm_code not in dist_case_type_col.tree_algorithm_codes:
                raise exc.UnauthorizedAuthError(
                    f"User {user.id} has no read access to tree algorithm {tree_algorithm_code}"
                )

            # Get genetic distance protocol
            genetic_distance_protocol: model.GeneticDistanceProtocol = (
                self.repository.crud(  # type: ignore[assignment]
                    uow,
                    user.id,
                    model.GeneticDistanceProtocol,
                    None,
                    dist_col.genetic_distance_protocol_id,
                    CrudOperation.READ_ONE,
                )
            )
            seqdb_seq_distance_protocol_id = (
                genetic_distance_protocol.seqdb_seq_distance_protocol_id
            )

            # Special case: zero case_ids
            if not case_ids:
                retval: model.PhylogeneticTree = self.app.handle(
                    command.RetrievePhylogeneticTreeBySequencesCommand(
                        user=user,
                        tree_algorithm_code=tree_algorithm_code,
                        seqdb_seq_distance_protocol_id=seqdb_seq_distance_protocol_id,
                        sequence_ids=[],
                    )
                )
                retval.genetic_distance_protocol_id = genetic_distance_protocol.id
                return retval

            # Create temporary case_abac only for this case type and the
            # seq_case_type_col_id having the same rights as the dist_case_type_col
            temp_case_abac = model.CaseAbac(
                is_full_access=case_abac.is_full_access,
                case_type_access_abacs={},
                case_type_share_abacs={},
            )
            for data_collection_id, x in case_abac.case_type_access_abacs.get(
                case_type_id, {}
            ).items():
                if dist_case_type_col_id not in x.read_case_type_col_ids:
                    continue
                if case_type_id not in temp_case_abac.case_type_access_abacs:
                    temp_case_abac.case_type_access_abacs[case_type_id] = {}
                temp_case_abac.case_type_access_abacs[case_type_id][
                    data_collection_id
                ] = model.CaseTypeAccessAbac(
                    read_case_type_col_ids={seq_case_type_col_id},
                    **x.model_dump(exclude={"read_case_type_col_ids"}),
                )

            # @ABAC: Get cases
            cases = self._retrieve_cases_with_content_right(
                uow,
                user.id,
                temp_case_abac,
                enum.CaseRight.READ_CASE,
                case_ids=case_ids,
                case_type_ids={case_type_id},
                filter_content=True,
            )

            # Get sequence_ids from seq_case_type_col
            case_sequence_map = {}
            for case in cases:
                sequence_id = case.content.get(seq_case_type_col_id)
                if sequence_id:
                    case_sequence_map[case.id] = UUID(sequence_id)

            # Retrieve tree and remove sequence_ids to avoid leaking information
            sequence_ids = list(case_sequence_map.values())
            sequence_case_map = {y: x for x, y in case_sequence_map.items()}
            phylogenetic_tree: model.PhylogeneticTree = self.app.handle(
                command.RetrievePhylogeneticTreeBySequencesCommand(
                    user=cmd.user,
                    tree_algorithm_code=tree_algorithm_code,
                    seqdb_seq_distance_protocol_id=seqdb_seq_distance_protocol_id,
                    sequence_ids=sequence_ids,
                    props={
                        "leaf_id_mapper": lambda x: sequence_case_map[x],
                    },
                )
            )
            phylogenetic_tree.genetic_distance_protocol_id = (
                genetic_distance_protocol.id
            )
            phylogenetic_tree.sequence_ids = None

        return phylogenetic_tree

    def retrieve_genetic_sequence(
        self,
        cmd: command.RetrieveGeneticSequenceByCaseCommand,
    ) -> list[model.GeneticSequence]:
        seq_case_type_col_id = cmd.genetic_sequence_case_type_col_id
        case_ids = cmd.case_ids
        user: model.User
        user, repository = self._get_user_and_repository(cmd)
        assert user.id is not None

        # Special case: zero case_ids
        if not case_ids:
            return []

        case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
        assert case_abac is not None

        with repository.uow() as uow:

            # @ABAC: Get cases and sequence_ids
            cases = self._retrieve_cases_with_content_right(
                uow,
                user.id,
                case_abac,
                enum.CaseRight.READ_CASE,
                case_ids=case_ids,
                filter_content=True,
            )
            seq_ids = [UUID(x.content.get(seq_case_type_col_id)) for x in cases]

            # Retrieve sequences
            genetic_sequences: list[model.GeneticSequence] = self.app.handle(
                command.RetrieveGeneticSequenceByIdCommand(
                    user=user,
                    seq_ids=seq_ids,
                )
            )

        return genetic_sequences

    def _crud_metadata(
        self,
        uow: BaseUnitOfWork,
        cmd: command.CrudCommand,
    ) -> Any:
        """Logic for handling metadata commands"""
        # Metadata admin or above: no @ABAC applied
        assert cmd.user
        if cmd.user.roles.intersection(enum.RoleSet.GE_METADATA_ADMIN.value):
            # Metadata admin and above have access to all metadata: no ABAC
            # applied, only RBAC
            return self._crud_metadata_by_admin(uow, cmd)
        return self._crud_metadata_by_non_admin(uow, cmd)

    def _crud_metadata_by_admin(
        self, uow: BaseUnitOfWork, cmd: command.CrudCommand
    ) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
        """Metadata admin command handling, no ABAC applied"""
        self._crud_cascade_delete(uow, cmd)
        return super().crud(cmd)  # type:ignore[return-value]

    def _crud_metadata_by_non_admin(
        self,
        uow: BaseUnitOfWork,
        cmd: command.CrudCommand,
    ) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
        """Metadata user command handling, ABAC applied"""
        # @ABAC: get case abac
        case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)

        # Special case: no policy, allows for internal commands to retrieve all
        if not case_abac:
            # No policy: allows for internal commands to retrieve all
            return super().crud(cmd)  # type:ignore[return-value]

        # Initialise some
        is_read = cmd.operation in CrudOperationSet.READ_OR_EXISTS.value
        is_delete = cmd.operation in CrudOperationSet.DELETE.value
        access_filter: Filter | None = None

        if not is_read:
            # Only read operations are allowed for metadata commands for these
            # users
            raise AssertionError("Unexpected operation")

        if isinstance(cmd, command.CaseTypeCrudCommand):
            case_abac
            valid_case_type_ids = case_abac.get_case_types_with_any_rights()
            access_filter = CaseService._compose_id_filter(("id", valid_case_type_ids))
            # No cascade delete to force conscious decision to delete from other models
            return self._crud_with_access_filter(uow, cmd, access_filter)

        elif isinstance(cmd, command.CaseTypeSetMemberCrudCommand):
            valid_case_type_ids = case_abac.get_case_types_with_any_rights()
            access_filter = CaseService._compose_id_filter(
                ("case_type_id", valid_case_type_ids)
            )
            return self._crud_with_access_filter(uow, cmd, access_filter)

        elif isinstance(cmd, command.CaseTypeSetCrudCommand):
            valid_case_type_ids = case_abac.get_case_types_with_any_rights()
            valid_case_type_set_ids: set[UUID] = (
                self._read_association_with_valid_ids(  # type:ignore[assignment]
                    command.CaseTypeSetMemberCrudCommand,
                    "case_type_set_id",
                    "case_type_id",
                    valid_ids2=valid_case_type_ids,
                    match_all2=is_delete,  # delete requires all case types
                    return_type="ids1",
                    uow=uow,
                    user=cmd.user,
                )
            )
            access_filter = CaseService._compose_id_filter(
                ("id", valid_case_type_set_ids)
            )
            # No cascade delete to force conscious decision to delete from other models
            return self._crud_with_access_filter(uow, cmd, access_filter)

        elif isinstance(cmd, command.CaseTypeColCrudCommand):
            valid_case_type_col_ids = case_abac.get_case_type_cols_with_any_rights()
            access_filter = CaseService._compose_id_filter(
                ("id", valid_case_type_col_ids)
            )
            # No cascade delete to force conscious decision to delete from other models
            return self._crud_with_access_filter(uow, cmd, access_filter)

        elif isinstance(cmd, command.CaseTypeColSetMemberCrudCommand):
            valid_case_type_col_ids = case_abac.get_case_type_cols_with_any_rights()
            access_filter = CaseService._compose_id_filter(
                ("case_type_col_id", valid_case_type_col_ids)
            )
            return self._crud_with_access_filter(uow, cmd, access_filter)

        elif isinstance(cmd, command.CaseTypeColSetCrudCommand):
            # Determine valid case type cols as those with any rights
            valid_case_type_col_ids = case_abac.get_case_type_cols_with_any_rights()
            valid_case_type_col_set_ids: set[UUID] = (
                self._read_association_with_valid_ids(  # type:ignore[assignment]
                    command.CaseTypeColSetMemberCrudCommand,
                    "case_type_col_set_id",
                    "case_type_col_id",
                    valid_ids2=valid_case_type_col_ids,
                    match_all2=is_delete,  # delete requires all case type cols
                    return_type="ids1",
                    uow=uow,
                    user=cmd.user,
                )
            )
            access_filter = CaseService._compose_id_filter(
                ("id", valid_case_type_col_set_ids)
            )
            # No cascade delete to force conscious decision to delete from other models
            return self._crud_with_access_filter(uow, cmd, access_filter)

        raise AssertionError("Unexpected operation")

    def _crud_data(
        self,
        uow: BaseUnitOfWork,
        cmd: command.CrudCommand,
    ) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
        """Logic for handling data commands"""
        assert cmd.user
        # App admin or above: no @ABAC applied
        if cmd.user.roles.intersection(enum.RoleSet.GE_APP_ADMIN.value):
            return self._crud_data_by_admin(uow, cmd)
        return self._crud_data_by_non_admin(uow, cmd)

    def _crud_data_by_admin(
        self,
        uow: BaseUnitOfWork,
        cmd: command.CrudCommand,
    ) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
        """Data admin command handling, no ABAC applied"""
        # Non-ABAC restrictions not enforced anywhere else
        is_create = cmd.operation in CrudOperationSet.CREATE.value
        is_update = cmd.operation in CrudOperationSet.UPDATE.value
        if (is_create or is_update) and isinstance(
            cmd, command.CaseSetMemberCrudCommand
        ):
            # Verify that the case set and case have the same case type
            self._verify_case_set_member_case_type(
                cmd.user, cmd.get_objs()  # type:ignore[arg-type]
            )

        # Any other operation
        self._crud_cascade_delete(uow, cmd)
        return super().crud(cmd)  # type:ignore[return-value]

    def _crud_data_by_non_admin(
        self,
        uow: BaseUnitOfWork,
        cmd: command.CrudCommand,
    ) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
        """Data user command handling, ABAC applied"""
        # @ABAC: get case abac
        case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)

        # Special case: no policy, allows for internal commands to retrieve all
        if case_abac is None:
            # No policy: allows for internal commands to retrieve all
            return super().crud(cmd)  # type:ignore[return-value]

        # Initialise some
        is_create = cmd.operation in CrudOperationSet.CREATE.value
        is_read = cmd.operation in CrudOperationSet.READ_OR_EXISTS.value
        is_read_all = cmd.operation == CrudOperation.READ_ALL
        is_update = cmd.operation in CrudOperationSet.UPDATE.value
        is_delete = cmd.operation in CrudOperationSet.DELETE.value
        is_delete_all = cmd.operation == CrudOperation.DELETE_ALL
        access_filter: Filter | None = None
        case_sets: list[model.CaseSet]
        cases: list[model.Case]
        assert cmd.user is not None and cmd.user.id is not None

        # Handle each type of command
        if isinstance(cmd, command.CaseSetCrudCommand):
            # Determine valid case types and data collections
            case_set_ids = cmd.get_obj_ids()
            if is_create:
                # Implemented through separate create case set command
                raise AssertionError("Unexpected operation")
            elif is_read:
                # At least one data collection with read access is required
                retval = self._retrieve_case_sets_with_content_right(  # type:ignore[return-value]
                    uow,
                    cmd.user.id,
                    case_abac,
                    enum.CaseRight.READ_CASE_SET,
                    case_set_ids=case_set_ids,  # type:ignore[arg-type]
                    filter=cmd.query_filter,
                )
                return (
                    retval[0] if cmd.operation == CrudOperation.READ_ONE else retval
                )  # type:ignore[arg-type]
            elif is_update:
                # At least one data collection with write access is required
                case_sets = self._retrieve_case_sets_with_content_right(  # type:ignore[return-value]
                    uow,
                    cmd.user.id,
                    case_abac,
                    enum.CaseRight.WRITE_CASE_SET,
                    case_set_ids=case_set_ids,  # type:ignore[arg-type]
                )
                return super().crud(cmd)
            elif is_delete:
                # All linked data collections have remove right
                if is_delete_all:
                    # Delete all not allowed due to potential large number of case sets
                    raise exc.UnauthorizedAuthError(
                        f"Operation {cmd.operation.value} not allowed for case sets for this user"
                    )
                # Get all case sets and data collection links
                assert case_set_ids is not None
                case_sets = self.repository.crud(  # type:ignore[assignment]
                    uow,
                    cmd.user.id,
                    model.CaseSet,
                    None,
                    case_set_ids,
                    CrudOperation.READ_SOME,
                )
                case_set_data_collection_map, _ = (
                    self._retrieve_case_set_data_collections_map(
                        uow,
                        cmd.user.id,
                        obj_ids1=case_set_ids,  # type:ignore[arg-type]
                    )
                )
                # Check if the user has access to all data collections of all requested
                # case sets
                for case_set in case_sets:
                    data_collection_ids = case_set_data_collection_map.get(
                        case_set.id, set()  # type:ignore[arg-type]
                    )
                    is_allowed = case_abac.is_allowed(
                        case_set.case_type_id,
                        enum.CaseRight.REMOVE_CASE_SET,
                        True,
                        created_in_data_collection_id=case_set.created_in_data_collection_id,
                        current_data_collection_ids=data_collection_ids,
                    )
                    if not is_allowed:
                        raise exc.UnauthorizedAuthError(
                            f"User {cmd.user.id} is not allowed to delete case set {case_set.id}"
                        )
                # Delete with cascade
                self._crud_cascade_delete(uow, cmd)
                return super().crud(cmd)  # type:ignore[return-value]
            else:
                raise AssertionError("Unexpected operation")

        elif isinstance(cmd, command.CaseCrudCommand):
            # Determine valid case types and data collections
            case_set_ids = cmd.get_obj_ids()
            if is_create | is_read | is_update:
                # Implemented through separate create case set command
                raise AssertionError("Unexpected operation")
            elif is_delete:
                # All linked data collections have remove right
                if is_delete_all:
                    # Delete all not allowed due to potential large number of case
                    raise exc.UnauthorizedAuthError(
                        f"Operation {cmd.operation.value} not allowed for cases for this user"
                    )
                # Get all cases and data collection links
                assert case_ids is not None
                cases = self.repository.crud(  # type:ignore[assignment]
                    uow,
                    cmd.user.id,
                    model.Case,
                    None,
                    case_ids,
                    CrudOperation.READ_SOME,
                )
                case_data_collection_map, _ = self._retrieve_case_data_collections_map(
                    uow,
                    cmd.user.id,
                    obj_ids1=case_ids,  # type:ignore[arg-type]
                )
                # Check if the user has access to all data collections of all requested
                # cases
                for case in cases:
                    data_collection_ids = case_data_collection_map.get(
                        case.id, set()  # type:ignore[arg-type]
                    )
                    is_allowed = case_abac.is_allowed(
                        case.case_type_id,
                        enum.CaseRight.REMOVE_CASE,
                        True,
                        created_in_data_collection_id=case.created_in_data_collection_id,
                        current_data_collection_ids=data_collection_ids,
                    )
                    if not is_allowed:
                        raise exc.UnauthorizedAuthError(
                            f"User {cmd.user.id} is not allowed to delete case {case.id}"
                        )
                # Delete with cascade
                self._crud_cascade_delete(uow, cmd)
                return super().crud(cmd)  # type:ignore[return-value]
            else:
                raise AssertionError("Unexpected operation")

        elif isinstance(cmd, command.CaseSetMemberCrudCommand):
            # Delete all not allowed due to potential large number of case set members
            if is_delete_all or is_update:
                raise exc.UnauthorizedAuthError(
                    f"Operation {cmd.operation.value} not allowed for case set members for this user"
                )

            # Get case set members
            case_set_members: list[model.CaseSetMember]
            if is_create:
                # Must be able to write the case set and read the case
                case_set_members = cmd.get_objs()  # type:ignore[assignment]
            elif is_read_all:
                # Must be able to read or write the case set and read the case
                case_set_members = self.repository.crud(  # type:ignore[assignment]
                    uow,
                    cmd.user.id,
                    model.CaseSetMember,
                    None,
                    None,
                    CrudOperation.READ_ALL,
                    filter=cmd.query_filter,
                )
            elif is_read or is_delete:
                # Must be able to read or write the case set and read the case
                case_set_members = self.repository.crud(  # type:ignore[assignment]
                    uow,
                    cmd.user.id,
                    model.CaseSetMember,
                    None,
                    cmd.get_obj_ids(),
                    CrudOperation.READ_SOME,
                )
            elif is_update:
                # Should not be allowed
                raise AssertionError("Update not allowed for case set members")
            else:
                raise AssertionError("Unexpected operation")

            # All operations require read access to the case: retrieve the cases while
            # checking for this read right to determine this
            cases = self._retrieve_cases_with_content_right(
                uow,
                cmd.user.id,
                case_abac,
                enum.CaseRight.READ_CASE,
                case_ids=list({x.case_id for x in case_set_members}),
                filter_content=False,
                on_invalid_case_id=(
                    "ignore" if is_read_all or is_delete_all else "raise"
                ),
            )

            # Retrieve the case sets while checking for the correct right(s)
            case_set_ids = {x.case_set_id for x in case_set_members}
            case_sets = self._retrieve_case_sets_with_content_right(
                uow,
                cmd.user.id,
                case_abac,
                enum.CaseRight.READ_CASE_SET,
                case_set_ids=list(case_set_ids),
                on_invalid_case_set_id="ignore",
            )
            if is_delete and not case_set_ids.issubset({x.id for x in case_sets}):
                # Also check the write case set right since not all case sets have the
                # read right
                case_sets += self._retrieve_case_sets_with_content_right(
                    uow,
                    cmd.user.id,
                    case_abac,
                    enum.CaseRight.WRITE_CASE_SET,
                    case_set_ids=list(case_set_ids),
                    on_invalid_case_set_id="ignore",
                )

            # Check if the user has access to all requested case sets
            unauthorized_case_set_ids = case_set_ids - {x.id for x in case_sets}
            if unauthorized_case_set_ids:
                if is_read_all:
                    # unauthorized case set ids not applicable, filter out the case set
                    # members instead
                    case_set_members = [
                        x
                        for x in case_set_members
                        if x.case_set_id not in unauthorized_case_set_ids
                    ]
                else:
                    unauthorized_case_set_ids_str = ", ".join(
                        [str(x) for x in unauthorized_case_set_ids]
                    )
                    raise exc.UnauthorizedAuthError(
                        f"User {cmd.user.id} does not have access to case set(s): {unauthorized_case_set_ids_str}"
                    )

            # Execute command in case of create or delete, return case set
            # members in case of read
            if is_create or is_delete:
                return super().crud(cmd)  # type:ignore[return-value]
            elif is_read:
                return (
                    case_set_members[0]  # type:ignore[return-value]
                    if cmd.operation == CrudOperation.READ_ONE
                    else case_set_members
                )
            else:
                raise AssertionError("Unexpected operation")

        elif isinstance(cmd, command.CaseDataCollectionLinkCrudCommand):
            # Read all without filter and delete all not allowed due to potential large
            # number of case data collection links
            if (is_read_all and not cmd.query_filter) or is_delete_all or is_update:
                raise exc.UnauthorizedAuthError(
                    f"Operation {cmd.operation.value} not allowed for case data collection links for this user"
                )

            # Get case data collection links
            case_data_collection_links: list[model.CaseDataCollectionLink]
            if is_create:
                # Must be able to add the case to all the data collections
                case_data_collection_links = cmd.get_objs()  # type:ignore[assignment]
            elif is_read_all:
                # Must be able to read or write the case in all the data collections
                case_data_collection_links = (
                    self.repository.crud(  # type:ignore[assignment]
                        uow,
                        cmd.user.id,
                        model.CaseDataCollectionLink,
                        None,
                        None,
                        CrudOperation.READ_ALL,
                        filter=cmd.query_filter,
                    )
                )
            elif is_read or is_delete:
                # Must be able to read or write the case in (for is_read), or remove from
                # (for is_delete), all the data collections
                case_data_collection_links = (
                    self.repository.crud(  # type:ignore[assignment]
                        uow,
                        cmd.user.id,
                        model.CaseDataCollectionLink,
                        None,
                        cmd.get_obj_ids(),
                        CrudOperation.READ_SOME,
                    )
                )
            elif is_update:
                # Should not be allowed
                raise AssertionError(
                    "Update not allowed for case data collection links"
                )
            else:
                raise AssertionError("Unexpected operation")

            # Go over each case and check if the user has the required rights to it
            case_data_collection_map: dict[UUID, set[UUID]] = (
                map_paired_elements(  # type:ignore[assignment]
                    (
                        (x.case_id, x.data_collection_id)
                        for x in case_data_collection_links
                    ),
                    as_set=True,
                )
            )
            case_ids = set(case_data_collection_map.keys())
            cases = self.repository.crud(  # type:ignore[assignment]
                uow,
                cmd.user.id,
                model.Case,
                None,
                list(case_ids),
                CrudOperation.READ_SOME,
            )
            if is_read:
                # Get read or write access per case type
                has_read_access = case_abac.get_combinations_with_access_right(
                    enum.CaseRight.READ_CASE
                )
                has_write_access = case_abac.get_combinations_with_access_right(
                    enum.CaseRight.WRITE_CASE
                )
                has_access = {x: set(y) for x, y in has_read_access.items()}
                for x, y in has_write_access.items():
                    if x in has_access:
                        has_access[x].update(y)
                    else:
                        has_access[x] = y
            for case in cases:
                case_type_id = case.case_type_id
                created_in_data_collection_id = case.created_in_data_collection_id
                data_collection_ids = case_data_collection_map[
                    case.id  # type:ignore[index]
                ]
                if is_read:
                    if not data_collection_ids.intersection(
                        has_access.get(case_type_id, set())
                    ):
                        raise exc.UnauthorizedAuthError(
                            f"User {cmd.user.id} does not have read or write access to case {case.id} in all data collections"
                        )
                elif is_create:
                    is_allowed = case_abac.is_allowed(
                        case_type_id,
                        enum.CaseRight.ADD_CASE,
                        False,
                        created_in_data_collection_id=created_in_data_collection_id,
                        current_data_collection_ids=data_collection_ids,
                        tgt_data_collection_ids=data_collection_ids,
                    )
                    if not is_allowed:
                        raise exc.UnauthorizedAuthError(
                            f"User {cmd.user.id} does not have add access to case {case.id} to all data collections"
                        )
                elif is_delete:
                    is_allowed = case_abac.is_allowed(
                        case_type_id,
                        enum.CaseRight.REMOVE_CASE,
                        False,
                        created_in_data_collection_id=created_in_data_collection_id,
                        current_data_collection_ids=data_collection_ids,
                        tgt_data_collection_ids=data_collection_ids,
                    )
                    if not is_allowed:
                        raise exc.UnauthorizedAuthError(
                            f"User {cmd.user.id} does not have remove access to case {case.id} to all data collections"
                        )

            # Execute command in case of create or delete, return case data collection
            # links in case of read
            if is_create or is_delete:
                return super().crud(cmd)  # type:ignore[return-value]
            elif is_read:
                return (
                    case_data_collection_links[0]  # type:ignore[return-value]
                    if cmd.operation == CrudOperation.READ_ONE
                    else case_data_collection_links
                )
            else:
                raise AssertionError("Unexpected operation")

        elif isinstance(cmd, command.CaseSetDataCollectionLinkCrudCommand):
            # Read all without filter and delete all not allowed due to potential large
            # number of case set data collection links
            if (is_read_all and not cmd.query_filter) or is_delete_all or is_update:
                raise exc.UnauthorizedAuthError(
                    f"Operation {cmd.operation.value} not allowed for case set data collection links for this user"
                )

            # Get case set data collection links
            case_set_data_collection_links: list[model.CaseSetDataCollectionLink]
            if is_create:
                # Must be able to add the case set to all the data collections
                case_set_data_collection_links = (
                    cmd.get_objs()  # type:ignore[assignment]
                )
            elif is_read_all:
                # Must be able to read or write the case set in all the data collections
                case_set_data_collection_links = (
                    self.repository.crud(  # type:ignore[assignment]
                        uow,
                        cmd.user.id,
                        model.CaseSetDataCollectionLink,
                        None,
                        None,
                        CrudOperation.READ_ALL,
                        filter=cmd.query_filter,
                    )
                )
            elif is_read or is_delete:
                # Must be able to read or write the case set in (for is_read), or remove from
                # (for is_delete), all the data collections
                case_set_data_collection_links = (
                    self.repository.crud(  # type:ignore[assignment]
                        uow,
                        cmd.user.id,
                        model.CaseSetDataCollectionLink,
                        None,
                        cmd.get_obj_ids(),
                        CrudOperation.READ_SOME,
                    )
                )
            elif is_update:
                # Should not be allowed
                raise AssertionError(
                    "Update not allowed for case set data collection links"
                )
            else:
                raise AssertionError("Unexpected operation")

            # Go over each case set and check if the user has the required rights to it
            case_set_data_collection_map: dict[UUID, set[UUID]] = (
                map_paired_elements(  # type:ignore[assignment]
                    (
                        (x.case_set_id, x.data_collection_id)
                        for x in case_set_data_collection_links
                    ),
                    as_set=True,
                )
            )
            case_set_ids = set(case_set_data_collection_map.keys())
            case_sets = self.repository.crud(  # type:ignore[assignment]
                uow,
                cmd.user.id,
                model.CaseSet,
                None,
                list(case_set_ids),
                CrudOperation.READ_SOME,
            )
            if is_read:
                # Get read or write access per case type
                has_read_access = case_abac.get_combinations_with_access_right(
                    enum.CaseRight.READ_CASE_SET
                )
                has_write_access = case_abac.get_combinations_with_access_right(
                    enum.CaseRight.WRITE_CASE_SET
                )
                has_access = {x: set(y) for x, y in has_read_access.items()}
                for x, y in has_write_access.items():
                    if x in has_access:
                        has_access[x].update(y)
                    else:
                        has_access[x] = y
            for case_set in case_sets:
                case_type_id = case_set.case_type_id
                created_in_data_collection_id = case_set.created_in_data_collection_id
                data_collection_ids = case_set_data_collection_map[
                    case_set.id  # type:ignore[index]
                ]
                if is_read:
                    if not data_collection_ids.intersection(
                        has_access.get(case_type_id, set())
                    ):
                        raise exc.UnauthorizedAuthError(
                            f"User {cmd.user.id} does not have read or write access to case set {case_set.id} in all data collections"
                        )
                elif is_create:
                    is_allowed = case_abac.is_allowed(
                        case_type_id,
                        enum.CaseRight.ADD_CASE_SET,
                        False,
                        created_in_data_collection_id=created_in_data_collection_id,
                        current_data_collection_ids=data_collection_ids,
                        tgt_data_collection_ids=data_collection_ids,
                    )
                    if not is_allowed:
                        raise exc.UnauthorizedAuthError(
                            f"User {cmd.user.id} does not have add access to case set {case_set.id} to all data collections"
                        )
                elif is_delete:
                    is_allowed = case_abac.is_allowed(
                        case_type_id,
                        enum.CaseRight.REMOVE_CASE_SET,
                        False,
                        created_in_data_collection_id=created_in_data_collection_id,
                        current_data_collection_ids=data_collection_ids,
                        tgt_data_collection_ids=data_collection_ids,
                    )
                    if not is_allowed:
                        raise exc.UnauthorizedAuthError(
                            f"User {cmd.user.id} does not have remove access to case set {case_set.id} to all data collections"
                        )

            # Execute command in case of create or delete, return case set data
            # collection links in case of read
            if is_create or is_delete:
                return super().crud(cmd)  # type:ignore[return-value]
            elif is_read:
                return (
                    case_set_data_collection_links[0]  # type:ignore[return-value]
                    if cmd.operation == CrudOperation.READ_ONE
                    else case_set_data_collection_links
                )
            else:
                raise AssertionError("Unexpected operation")

        raise AssertionError("Unexpected operation")

    def _crud_cascade_delete(
        self, uow: BaseUnitOfWork, cmd: command.CrudCommand
    ) -> None:
        """
        In case of a delete operation, cascade delete all instances of any
        linked_model_classes that are linked to the instances in cmd.
        """
        is_delete = cmd.operation in CrudOperationSet.DELETE.value
        if not is_delete:
            # Not a delete opertion: nothing to do
            return
        model_class = cmd.MODEL_CLASS
        link_model_classes: list[Type[model.Model]] | None = None
        for (
            model_base_class,
            link_model_classes_tuple,
        ) in BaseCaseService.CASCADE_DELETE_MODEL_CLASSES.items():
            if issubclass(model_class, model_base_class):
                link_model_classes = list(link_model_classes_tuple)
                break
        if link_model_classes is None:
            # No cascade delete: nothing to do
            return
        assert cmd.user is not None and cmd.user.id is not None
        obj_ids: set[UUID] | None = cmd.get_obj_ids(as_set=True)
        # Go over each link_model_class and delete all instances that are linked to
        # the instances in cmd
        for link_model_class in link_model_classes:
            assert link_model_class.ENTITY is not None
            get_obj_id = link_model_class.ENTITY.get_obj_id
            if cmd.operation == CrudOperation.DELETE_ALL:
                # Special case: delete all instances
                self.repository.crud(
                    uow,
                    cmd.user.id,
                    link_model_class,
                    None,
                    None,
                    CrudOperation.DELETE_ALL,
                )
                continue
            assert obj_ids is not None
            # Get the instances that are linked to the instances in cmd
            get_link_id = link_model_class.ENTITY.get_link_id(model_class)
            link_objs: list = self.repository.crud(  # type:ignore[assignment]
                uow,
                cmd.user.id,
                link_model_class,
                None,
                None,
                CrudOperation.READ_ALL,
            )
            link_obj_ids = [
                get_obj_id(x) for x in link_objs if get_link_id(x) in obj_ids
            ]
            # Delete these instances
            self.repository.crud(
                uow,
                cmd.user.id,
                link_model_class,
                None,
                link_obj_ids,
                CrudOperation.DELETE_SOME,
            )

    def _crud_with_access_filter(
        self,
        uow: BaseUnitOfWork,
        cmd: command.CrudCommand,
        access_filter: Filter | None = None,
        cascade_if_delete: bool = False,
    ) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
        # Set access filter if any and call generic crud
        orig_access_filter = cmd.access_filter
        if access_filter:
            if cmd.access_filter:
                cmd.access_filter = CompositeFilter(
                    filters=[access_filter, cmd.access_filter],  # type: ignore[list-item]
                    operator=BooleanOperator.AND,
                )
            else:
                cmd.access_filter = access_filter
        if cascade_if_delete:
            self._crud_cascade_delete(uow, cmd)
        retval = super().crud(cmd)
        cmd.access_filter = orig_access_filter
        return retval  # type:ignore[return-value]

    def _read_association_with_valid_ids(
        self,
        command_class: Type[command.CrudCommand],
        field_name1: str,
        field_name2: str,
        valid_ids1: set[UUID] | frozenset[UUID] | None = None,
        valid_ids2: set[UUID] | frozenset[UUID] | None = None,
        match_all1: bool = False,
        match_all2: bool = False,
        return_type: str = "objects",
        uow: BaseUnitOfWork | None = None,
        user: model.User | None = None,
    ) -> list[model.Model] | list[UUID] | dict[UUID, set[UUID]]:
        # TODO: this can be a generic service/repository method (ids should be Hashable instead of UUID)
        # Parse arguments
        if return_type not in {"objects", "ids1", "ids2", "id_map12", "id_map21"}:
            raise ValueError(f"Invalid return_type: {return_type}")
        if match_all1 and match_all2:
            raise ValueError("match_all1 and match_all2 cannot both be True")
        id_map12 = return_type == "id_map12"
        id_map21 = return_type == "id_map21"
        if id_map12 and match_all1:
            raise ValueError("match_all1 must be False if id_map12 is True")
        if id_map21 and match_all2:
            raise ValueError("match_all2 must be False if id_map21 is True")
        if return_type == "ids1" and match_all1:
            raise ValueError("match_all1 must be False if return_type is ids1")
        if return_type == "ids2" and match_all2:
            raise ValueError("match_all2 must be False if return_type is ids2")
        # Create filter
        filter: Filter | None
        if valid_ids1 is not None:
            if not isinstance(valid_ids1, frozenset):
                valid_ids1 = frozenset(valid_ids1)
            if not valid_ids1:
                # Empty set of valid values -> no matches
                if return_type in {"id_map12", "id_map21"}:
                    return dict()
                return []
            if valid_ids2 is not None:
                if not valid_ids2:
                    # Empty set of valid values -> no matches
                    if return_type in {"id_map12", "id_map21"}:
                        return dict()
                    return []
                if not isinstance(valid_ids2, frozenset):
                    valid_ids2 = frozenset(valid_ids2)
                filter = CompositeFilter(
                    filters=[
                        UuidSetFilter(key=field_name1, members=valid_ids1),
                        UuidSetFilter(key=field_name2, members=valid_ids2),
                    ],
                    operator=BooleanOperator.AND,
                )
            else:
                if match_all2:
                    raise ValueError("match_all2 must be False if valid_ids2 is None")
                if not isinstance(valid_ids1, frozenset):
                    valid_ids2 = frozenset(valid_ids2)
                filter = UuidSetFilter(key=field_name1, members=valid_ids1)
        elif valid_ids2 is not None:
            if not valid_ids2:
                # Empty set of valid values -> no matches
                if return_type in {"id_map12", "id_map21"}:
                    return dict()
                return []
            if match_all1:
                raise ValueError("match_all1 must be False if valid_ids1 is None")
            if not isinstance(valid_ids2, frozenset):
                valid_ids2 = frozenset(valid_ids2)
            filter = UuidSetFilter(key=field_name2, members=valid_ids2)
        else:
            if match_all1 or match_all2:
                raise ValueError(
                    "match_all1 and match_all2 must be False if valid_ids1 and valid_ids2 are None"
                )
            filter = None
        # Query repository
        cmd = command_class(
            user=user, operation=CrudOperation.READ_ALL, query_filter=filter
        )
        objs: list[model.Model]
        if uow:
            objs = self.crud_repository(uow, cmd)  # type: ignore[assignment]
        else:
            with self.repository.uow() as uow:
                objs = self.crud_repository(uow, cmd)  # type: ignore[assignment]
        ids1 = [getattr(x, field_name1) for x in objs]
        ids2 = [getattr(x, field_name2) for x in objs]
        # Apply id_map12/id_map21 and match_all1/match_all2 if necessary
        if id_map12 or id_map21 or match_all1 or match_all2:
            id_map: dict[UUID, set[UUID]] = {}
            if id_map12 or match_all2:
                # Create dict[id1, set[id2]]
                for id1, id2 in zip(ids1, ids2):
                    if id1 in id_map:
                        id_map[id1].add(id2)
                    else:
                        id_map[id1] = {id2}
                if match_all2:
                    # Keep only ids1 linked to all valid ids2
                    id_map = {
                        x: y for x, y in id_map.items() if len(y) == len(valid_ids2)  # type: ignore[arg-type]
                    }
                    if id_map12:
                        return id_map
                    elif return_type == "objects":
                        return [x for x, y in zip(objs, ids1) if y in id_map]
                    elif return_type == "ids1":
                        return list(id_map.keys())
                elif id_map12:
                    return id_map
                else:
                    raise AssertionError("Unexpected case")
            elif id_map21 or match_all1:
                # Create dict[id2, set[id1]]
                for id1, id2 in zip(ids1, ids2):
                    if id2 in id_map:
                        id_map[id2].add(id1)
                    else:
                        id_map[id2] = {id1}
                if match_all1:
                    # Keep only ids2 linked to all valid ids1
                    id_map = {
                        x: y for x, y in id_map.items() if len(y) == len(valid_ids1)  # type: ignore[arg-type]
                    }
                    if id_map21:
                        return id_map
                    elif return_type == "objects":
                        return [x for x, y in zip(objs, ids2) if y in id_map]
                    elif return_type == "ids2":
                        return list(id_map.keys())
                elif id_map21:
                    return id_map
                else:
                    raise AssertionError("Unexpected case")
            else:
                raise AssertionError("Unexpected case")
        # Return objs or ids for remaining cases
        if return_type == "objects":
            return objs
        if return_type == "ids1":
            return ids1
        if return_type == "ids2":
            return ids2
        raise AssertionError(f"Unexpected return_type: {return_type}")

    def _retrieve_case_sets_with_content_right(
        self,
        uow: BaseUnitOfWork,
        user_id: UUID,
        case_abac: model.CaseAbac,
        right: enum.CaseRight,
        case_set_ids: list[UUID] | None = None,
        case_type_ids: set[UUID] | None = None,
        filter: Filter | None = None,
        on_invalid_case_set_id: str = "raise",
    ) -> list[model.CaseSet]:
        # TODO: This is a temporary implementation, to be replaced by optimized query
        if right not in enum.CaseRightSet.CASE_SET_CONTENT.value:
            raise exc.InvalidArgumentsError(f"Invalid case abac right: {right.value}")
        if on_invalid_case_set_id not in {"raise", "ignore"}:
            raise exc.InvalidArgumentsError(
                f"Invalid on_invalid_case_set_id: {on_invalid_case_set_id}"
            )

        # Retrieve all case sets, potentially filtered
        case_sets: list[model.CaseSet]
        if case_set_ids:
            if filter:
                raise exc.InvalidArgumentsError(
                    "Cannot use datetime range filter with case set ids"
                )
            case_sets = self.repository.crud(  # type:ignore[assignment]
                uow,
                user_id,
                model.CaseSet,
                None,
                case_set_ids,
                CrudOperation.READ_SOME,
            )
        else:
            case_sets = self.repository.crud(  # type:ignore[assignment]
                uow,
                user_id,
                model.CaseSet,
                None,
                None,
                CrudOperation.READ_ALL,
                filter=filter,
            )

        # Filter on case_type_ids if any or verify that all case sets have a valid
        # case_type_id if case_set_ids is given
        # TODO: add more efficient implementation by adding this as a filter in the
        # call to the repository
        if case_type_ids is not None:
            if case_set_ids:
                if on_invalid_case_set_id == "raise":
                    if not all(x.case_type_id in case_type_ids for x in case_sets):
                        raise exc.InvalidArgumentsError(
                            f"Some case sets have invalid case type ids: {case_set_ids}"
                        )
                elif on_invalid_case_set_id == "ignore":
                    pass
                else:
                    raise AssertionError(
                        f"Invalid on_invalid_case_set_id: {on_invalid_case_set_id}"
                    )
            case_sets = [x for x in case_sets if x.case_type_id in case_type_ids]

        # Special case: full_access
        if case_abac.is_full_access:
            return case_sets

        # @ABAC: filter case sets to which the user has read access
        case_set_data_collections, _ = self._retrieve_case_set_data_collections_map(
            uow, user_id
        )
        has_access = case_abac.get_combinations_with_access_right(right)
        filtered_case_sets = []
        for case_set in case_sets:
            # Check if user has any access to case
            case_type_id = case_set.case_type_id
            if case_type_id not in has_access:
                if case_set_ids:
                    if on_invalid_case_set_id == "raise":
                        raise exc.UnauthorizedAuthError(
                            f"User {user_id} has no access to some requested cases"
                        )
                    elif on_invalid_case_set_id == "ignore":
                        pass
                    else:
                        raise AssertionError(
                            f"Invalid on_invalid_case_id: {on_invalid_case_set_id}"
                        )
                continue
            # Check if user has access to any of the data collections of the case set
            data_collection_ids = case_set_data_collections[
                case_set.id  # type:ignore[index]
            ]
            if not data_collection_ids.intersection(has_access[case_type_id]):
                if case_set_ids:
                    if on_invalid_case_set_id == "raise":
                        raise exc.UnauthorizedAuthError(
                            f"User {user_id} has no access to some requested case sets"
                        )
                    elif on_invalid_case_set_id == "ignore":
                        pass
                    else:
                        raise AssertionError(
                            f"Invalid on_invalid_case_set_id: {on_invalid_case_set_id}"
                        )
                continue
            # Keep case
            filtered_case_sets.append(case_set)
        return filtered_case_sets

    def _retrieve_cases_with_content_right(
        self,
        uow: BaseUnitOfWork,
        user_id: UUID,
        case_abac: model.CaseAbac,
        right: enum.CaseRight,
        case_ids: list[UUID] | None = None,
        case_type_ids: set[UUID] | None = None,
        datetime_range_filter: DatetimeRangeFilter | None = None,
        on_invalid_case_id: str = "raise",
        filter_content: bool = True,
        extra_access_case_type_col_ids: set[UUID] | None = None,
    ) -> list[model.Case]:
        # TODO: This is a temporary implementation, to be replaced by optimized query
        if right not in enum.CaseRightSet.CASE_CONTENT.value:
            raise exc.InvalidArgumentsError(f"Invalid case abac right: {right.value}")
        if on_invalid_case_id not in {"raise", "ignore"}:
            raise exc.InvalidArgumentsError(
                f"Invalid on_invalid_case_id: {on_invalid_case_id}"
            )

        # Retrieve all cases, potentially filtered by datetime range
        if datetime_range_filter:
            if datetime_range_filter.key and datetime_range_filter.key != "case_date":
                raise exc.InvalidArgumentsError(
                    f"Invalid datetime range filter key: {datetime_range_filter.key}"
                )
            datetime_range_filter.key = "case_date"
        cases: list[model.Case]
        if case_ids:
            if datetime_range_filter:
                raise exc.InvalidArgumentsError(
                    "Cannot use datetime range filter with case ids"
                )
            cases = self.repository.crud(  # type:ignore[assignment]
                uow,
                user_id,
                model.Case,
                None,
                case_ids,
                CrudOperation.READ_SOME,
            )
        else:
            cases = self.repository.crud(  # type:ignore[assignment]
                uow,
                user_id,
                model.Case,
                None,
                None,
                CrudOperation.READ_ALL,
                filter=datetime_range_filter,
            )

        # Filter on case_type_ids if any or verify that all cases have a valid
        # case_type_id if case_ids is given
        # TODO: add more efficient implementation by adding this as a filter in the
        # call to the repository
        if case_type_ids is not None:
            if case_ids:
                if not all(x.case_type_id in case_type_ids for x in cases):
                    raise exc.InvalidArgumentsError(
                        f"Some cases have invalid case type ids: {case_ids}"
                    )
                if on_invalid_case_id == "raise":
                    if not all(x.case_type_id in case_type_ids for x in cases):
                        raise exc.InvalidArgumentsError(
                            f"Some cases have invalid case type ids: {case_ids}"
                        )
                elif on_invalid_case_id == "ignore":
                    pass
                else:
                    raise AssertionError(
                        f"Invalid on_invalid_case_id: {on_invalid_case_id}"
                    )
            cases = [x for x in cases if x.case_type_id in case_type_ids]

        # Special case: full_access
        if case_abac.is_full_access:
            return cases

        # @ABAC: filter cases to which the user has read access, and optionally also
        # the content (case type cols)
        case_data_collections, _ = self._retrieve_case_data_collections_map(
            uow, user_id
        )
        has_access = case_abac.get_combinations_with_access_right(right)
        filtered_cases = []
        for case in cases:
            # Check if user has any access to case
            case_type_id = case.case_type_id
            if case_type_id not in has_access:
                if case_ids:
                    if on_invalid_case_id == "raise":
                        raise exc.UnauthorizedAuthError(
                            f"User {user_id} has no access to some requested cases"
                        )
                    elif on_invalid_case_id == "ignore":
                        pass
                    else:
                        raise AssertionError(
                            f"Invalid on_invalid_case_id: {on_invalid_case_id}"
                        )
                continue
            # Check if user has access to any data collection of the case
            data_collection_ids = case_data_collections[case.id]  # type:ignore[index]
            if not data_collection_ids.intersection(has_access[case_type_id]):
                if case_ids:
                    if on_invalid_case_id == "raise":
                        raise exc.UnauthorizedAuthError(
                            f"User {user_id} has no access to some requested cases"
                        )
                    elif on_invalid_case_id == "ignore":
                        pass
                    else:
                        raise AssertionError(
                            f"Invalid on_invalid_case_id: {on_invalid_case_id}"
                        )
                continue
            # Keep case
            filtered_cases.append(case)
            # Continue to next case if case content need not be filtered
            if not filter_content:
                continue
            # Determine which case type cols the user has access to
            data_collection_col_access = case_abac.case_type_access_abacs[case_type_id]
            case_type_col_ids = set()
            for data_collection_id in data_collection_ids:
                # Add case type cols with access to the case for this data
                # collection
                case_type_access_abac = data_collection_col_access.get(
                    data_collection_id
                )
                if case_type_access_abac is not None:
                    case_type_col_ids.update(
                        case_type_access_abac.read_case_type_col_ids
                    )
            if extra_access_case_type_col_ids is not None:
                case_type_col_ids.update(extra_access_case_type_col_ids)
            if not case_type_col_ids:
                data_collection_ids_str = ", ".join(
                    [str(x) for x in data_collection_ids]
                )
                raise AssertionError(
                    f"User {user_id} has zero columns with {right.value} access to case {case.id}, data collections ({data_collection_ids_str}) even though the case has some {right.value} access"
                )
            # Filter case content
            case.content = {
                x: y for x, y in case.content.items() if x in case_type_col_ids
            }
        return filtered_cases

    def _retrieve_case_data_collections_map(
        self, uow: BaseUnitOfWork, user_id: UUID, **kwargs: dict
    ) -> tuple[dict[UUID, set[UUID]], list[model.CaseDataCollectionLink]]:
        return self._retrieve_association_map(  # type:ignore[return-value]
            uow,
            user_id,
            model.CaseDataCollectionLink,
            "case_id",
            "data_collection_id",
            **kwargs,
        )

    def _retrieve_case_set_data_collections_map(
        self, uow: BaseUnitOfWork, user_id: UUID, **kwargs: dict
    ) -> tuple[dict[UUID, set[UUID]], list[model.CaseSetDataCollectionLink]]:
        return self._retrieve_association_map(  # type:ignore[return-value]
            uow,
            user_id,
            model.CaseSetDataCollectionLink,
            "case_set_id",
            "data_collection_id",
            **kwargs,
        )

    def _retrieve_case_case_sets_map(
        self, uow: BaseUnitOfWork, user_id: UUID, **kwargs: dict
    ) -> tuple[dict[UUID, set[UUID]], list[model.CaseSetMember]]:
        return self._retrieve_association_map(  # type:ignore[return-value]
            uow,
            user_id,
            model.CaseSetMember,
            "case_id",
            "case_set_id",
            **kwargs,
        )

    def _retrieve_association_map(
        self,
        uow: BaseUnitOfWork,
        user_id: UUID | None,
        association_class: Type[model.Model],
        link_field_name1: str,
        link_field_name2: str,
        **kwargs: dict,
    ) -> tuple[dict[UUID, set[UUID]], list[model.Model]]:
        """
        Get a dict[obj_id1, set[obj_ids]] based on the association stored in the association_class objs.
        """
        obj_ids1: frozenset[UUID] | None = kwargs.pop(  # type:ignore[assignment]
            "obj_ids1", None
        )
        obj_ids2: frozenset[UUID] | None = kwargs.pop(  # type:ignore[assignment]
            "obj_ids2", None
        )
        # Create a filter to restrict the association objs if necessary
        filter: Filter | None
        if obj_ids1:
            filter1 = UuidSetFilter(key=link_field_name1, members=obj_ids1)
        else:
            filter1 = None
        if obj_ids2:
            filter2 = UuidSetFilter(key=link_field_name2, members=obj_ids2)
        else:
            filter2 = None
        if filter1 and filter2:
            filter = CompositeFilter(
                filters=[filter1, filter2], operator=BooleanOperator.AND
            )
        elif filter1:
            filter = filter1
        elif filter2:
            filter = filter2
        else:
            filter = None
        # Retrieve association objs and convert to map
        association_objs: list = self.repository.crud(  # type:ignore[assignment]
            uow,
            user_id,
            association_class,
            None,
            None,
            CrudOperation.READ_ALL,
            filter=filter,
        )
        association_map: dict[UUID, set[UUID]] = (
            map_paired_elements(  # type:ignore[assignment]
                (
                    (getattr(x, link_field_name1), getattr(x, link_field_name2))
                    for x in association_objs
                ),
                as_set=True,
            )
        )

        return association_map, association_objs

    def _retrieve_sequence_column_data(
        self, uow: BaseUnitOfWork, user: model.User, seq_case_type_col_id: UUID
    ) -> tuple[model.CaseTypeCol, model.Col]:
        repository = self.repository
        seq_case_type_col: model.CaseTypeCol = repository.crud(  # type: ignore[assignment]
            uow,
            user.id,
            model.CaseTypeCol,
            None,
            seq_case_type_col_id,
            CrudOperation.READ_ONE,
        )
        seq_col: model.Col = repository.crud(  # type: ignore[assignment]
            uow,
            user.id,
            model.Col,
            None,
            seq_case_type_col.col_id,
            CrudOperation.READ_ONE,
        )
        if seq_col.col_type != enum.ColType.GENETIC_SEQUENCE:
            raise exc.InvalidArgumentsError(
                f"Case type column {seq_col.id} is not of type {enum.ColType.GENETIC_SEQUENCE.value}"
            )
        return seq_case_type_col, seq_col

    def _verify_case_filter(
        self, uow: BaseUnitOfWork, user: model.User, filter: CompositeFilter
    ) -> list[model.Col]:
        # Retrieve case type cols corresponding to filter keys
        filter_case_type_col_ids = filter.get_keys()
        filter_case_type_cols: list[model.CaseTypeCol] = (
            self.repository.crud(  # type:ignore[assignment]
                uow,
                user.id,
                model.CaseTypeCol,
                None,
                filter_case_type_col_ids,
                CrudOperation.READ_SOME,
            )
        )
        # Retrieve cols for case type cols
        cols: list[model.Col] = self.repository.crud(  # type:ignore[assignment]
            uow,
            user.id,
            model.Col,
            None,
            list(
                set(x.col_id for x in filter_case_type_cols)
            ),  # TODO: consider READ_SOME allowing duplicate ids
            CrudOperation.READ_SOME,
        )
        cols_ = {x.id: x for x in cols}
        cols = [cols_[x.col_id] for x in filter_case_type_cols]
        # Verify filter validity
        concept_valid_values = {}
        region_valid_values = {}
        for case_type_col, col, filter in zip(  # type:ignore[assignment]
            filter_case_type_cols, cols, filter.filters
        ):
            if col.concept_set_id or col.region_set_id:
                if isinstance(filter, StringSetFilter):
                    if col.concept_set_id:
                        # Get valid region set values
                        if col.concept_set_id not in concept_valid_values:
                            concept_set_members: list[model.ConceptSetMember] = (
                                self.app.handle(
                                    command.ConceptSetMemberCrudCommand(
                                        user=user,
                                        operation=CrudOperation.READ_ALL,
                                        query_filter=UuidSetFilter(
                                            key="concept_set_id",
                                            members={col.concept_set_id},  # type: ignore[arg-type]
                                        ),
                                    )
                                )
                            )
                            concepts = self.app.handle(
                                command.ConceptCrudCommand(
                                    user=user,
                                    operation=CrudOperation.READ_SOME,
                                    obj_ids=[x.concept_id for x in concept_set_members],
                                )
                            )
                            concept_valid_values[col.concept_set_id] = set(
                                [str(x.id).lower() for x in concepts]
                            )
                        valid_values = concept_valid_values[col.concept_set_id]
                    elif col.region_set_id:
                        # Get valid region set values
                        if col.region_set_id not in region_valid_values:
                            regions: list[model.Region] = self.app.handle(
                                command.RegionCrudCommand(
                                    user=user,
                                    operation=CrudOperation.READ_ALL,
                                    query_filter=UuidSetFilter(
                                        key="region_set_id",
                                        members={col.region_set_id},  # type: ignore[arg-type]
                                    ),
                                )
                            )
                            region_valid_values[col.region_set_id] = set(
                                [str(x.id).lower() for x in regions]
                            )
                        valid_values = region_valid_values[col.region_set_id]
                    # Handle invalid values
                    invalid_values = [
                        str(x)
                        for x in filter.members
                        if str(x).lower() not in valid_values
                    ]
                    if len(invalid_values):
                        invalid_values_str = ", ".join(invalid_values)
                        raise exc.InvalidArgumentsError(
                            f"Column {case_type_col.id}: invalid {filter.__class__.__name__} filter members: {invalid_values_str}"
                        )
                else:
                    raise exc.InvalidArgumentsError(
                        f"Column {case_type_col.id}: invalid filter type: {filter.__class__.__name__}"
                    )

        return cols

    def _verify_case_set_member_case_type(
        self, user: model.User, case_set_members: list[model.CaseSetMember]
    ) -> None:
        with self.repository.uow() as uow:
            case_set_ids = {x.case_set_id for x in case_set_members}
            case_ids = {x.case_id for x in case_set_members}
            case_sets_: list[model.CaseSet] = (
                self.repository.crud(  # type:ignore[assignment]
                    uow,
                    user.id if user else None,
                    model.CaseSet,
                    None,
                    list(case_set_ids),
                    CrudOperation.READ_SOME,
                )
            )
            case_sets = {x.id: x for x in case_sets_}
            cases_: list[model.Case] = self.repository.crud(  # type:ignore[assignment]
                uow,
                user.id if user else None,
                model.Case,
                None,
                list(case_ids),
                CrudOperation.READ_SOME,
            )
            cases = {x.id: x for x in cases_}
        invalid_case_set_member_ids = [
            x.id
            for x in case_set_members
            if case_sets[x.case_set_id].case_type_id != cases[x.case_id].case_type_id
        ]
        if invalid_case_set_member_ids:
            invalid_case_set_member_ids_str = ", ".join(
                [str(x) for x in invalid_case_set_member_ids]
            )
            raise exc.InvalidArgumentsError(
                f"Case set members invalid, case set and case must have the same case type: {invalid_case_set_member_ids_str}"
            )

    @staticmethod
    def _get_map_funs_for_filters(
        cols: Iterable[model.Col],
    ) -> list[Callable[[Any], Any]]:

        # Check validity of filter and generate map_funs
        map_funs = []
        for col in cols:
            if col.col_type == enum.ColType.TIME_DAY:
                map_funs.append(
                    lambda x: (
                        datetime.date.fromisoformat(x) if isinstance(x, str) else x
                    )
                )
            elif col.col_type in {
                enum.ColType.TIME_WEEK,
                enum.ColType.TIME_MONTH,
                enum.ColType.TIME_QUARTER,
                enum.ColType.TIME_YEAR,
                enum.ColType.GEO_REGION,
                enum.ColType.NOMINAL,
                enum.ColType.ORDINAL,
                enum.ColType.INTERVAL,
                enum.ColType.TEXT,
                enum.ColType.ID_DIRECT,
                enum.ColType.ID_PSEUDONYMISED,
                enum.ColType.ORGANIZATION,
                enum.ColType.OTHER,
            }:
                map_funs.append(lambda x: x if isinstance(x, str) else str(x))
            elif col.col_type == enum.ColType.DECIMAL_0:
                map_funs.append(lambda x: int(x) if isinstance(x, str) else x)
            elif col.col_type in {
                enum.ColType.DECIMAL_1,
                enum.ColType.DECIMAL_2,
                enum.ColType.DECIMAL_3,
                enum.ColType.DECIMAL_4,
                enum.ColType.DECIMAL_5,
                enum.ColType.DECIMAL_6,
            }:
                map_funs.append(lambda x: Decimal(x) if isinstance(x, str) else x)
            elif col.col_type == enum.ColType.GEO_LATLON:
                map_funs.append(
                    lambda x: (
                        (float(x.split(",")[0]), float(x.split(",")[1]))
                        if isinstance(x, str)
                        else x
                    )
                )
            else:
                raise exc.InvalidArgumentsError(
                    f"Unsupported column type: {col.col_type}"
                )
        return map_funs

    @staticmethod
    def _compose_id_filter(*key_and_ids: tuple[str, set[UUID]]) -> Filter:
        if len(key_and_ids) == 1:
            key, ids = key_and_ids[0]
            return UuidSetFilter(key=key, members=ids)  # type: ignore[arg-type]
        return CompositeFilter(
            filters=[
                UuidSetFilter(key=key, members=ids)  # type: ignore[arg-type]
                for key, ids in key_and_ids
            ],
            operator=BooleanOperator.AND,
        )



from gen_epix.casedb.domain import command, enum, model
from gen_epix.casedb.domain.policy.abac import BaseUpdateUserPolicy
from gen_epix.fastapp import Command


class UpdateUserPolicy(BaseUpdateUserPolicy):
    def is_allowed(self, cmd: Command) -> bool:
        user: model.User | None = cmd.user  # type: ignore[assignment]
        if user is None or user.id is None:
            return False
        roles = user.roles
        is_root = enum.Role.ROOT in roles

        if isinstance(cmd, command.InviteUserCommand):
            if user.email == cmd.email:
                # User cannot invite themselves
                return False
            tgt_user = model.User(**cmd.model_dump())
            tgt_roles_union = set(tgt_user.roles)
            is_organization_update = False
        elif isinstance(cmd, command.UpdateUserCommand):
            tgt_user = self.abac_service.app.user_manager.retrieve_user_by_id(  # type: ignore[assignment]
                cmd.tgt_user_id
            )
            tgt_roles_union = set(tgt_user.roles)
            if cmd.roles is not None:
                tgt_roles_union.update(cmd.roles)
            is_organization_update = (
                cmd.organization_id is not None
                and cmd.organization_id != tgt_user.organization_id
            )
        else:
            raise NotImplementedError

        # Handle ROOT, >=APP_ADMIN, <ORG_ADMIN
        if is_root:
            # Root user can invite/update anyone with any permissions
            return True
        tgt_user.roles = tgt_roles_union
        if len(roles.intersection(enum.RoleSet.GE_APP_ADMIN.value)) > 0:
            # APP_ADMIN users and above can invite/update anyone with less permissions
            # (so not another APP_ADMIN), and only if the new set of permissions is also
            # less than their own permissions
            return self._has_more_permissions(user, tgt_user)
        if not roles.intersection(enum.RoleSet.GE_ORG_ADMIN.value):
            # Only ORG_ADMIN users and above can invite/update users
            return False

        # ORG_ADMIN users can invite/update users of the organization(s) they are admin
        # of, but only if they have less permissions (so not another ORG_ADMIN), and
        # only if the new set of permissions is also less than their own permissions.
        # They cannot update the organization.
        if is_organization_update:
            # User cannot update the organization
            return False
        user_admin_organization_ids = self.abac_service.get_organizations_under_admin(
            user
        )
        if tgt_user.organization_id not in user_admin_organization_ids:
            # User is not admin of the organization(s) the target user is part of
            return False
        return self._has_more_permissions(user, tgt_user)

    def _has_more_permissions(self, user: model.User, tgt_user: model.User) -> bool:
        permissions = self.abac_service.app.user_manager.retrieve_user_permissions(user)  # type: ignore[arg-type]
        tgt_permissions = self.abac_service.app.user_manager.retrieve_user_permissions(
            tgt_user  # type: ignore[arg-type]
        )
        return tgt_permissions.issubset(permissions) and len(permissions) > len(
            tgt_permissions
        )



from __future__ import annotations

import logging
import uuid
from typing import Callable, Hashable
from uuid import UUID

from gen_epix.fastapp import App, Command
from gen_epix.omopdb.domain import enum, model
from gen_epix.omopdb.domain.service import BaseRbacService


class RbacService(BaseRbacService):

    def __init__(
        self,
        app: App,
        logger: logging.Logger | None = None,
        **kwargs: dict,
    ):
        kwargs["id_factory"] = kwargs.get("id_factory", uuid.uuid4)
        kwargs["service_type"] = kwargs.get(
            "service_type", BaseRbacService.SERVICE_TYPE
        )
        super().__init__(app, logger=logger, **kwargs)
        self._id_factory: Callable[[], UUID]
        # Register permissions without RBAC
        for (
            command_class,
            permission_type,
        ) in BaseRbacService.NO_RBAC_PERMISSIONS:
            permission = self.app.domain.get_permission(command_class, permission_type)
            self.register_permission_without_rbac(permission)

    def register_policies(self) -> None:
        self.register_rbac_policies()

    def retrieve_user_roles(self, user: model.User) -> set[Hashable]:
        return user.roles

    def retrieve_user_is_non_rbac_authorized(self, cmd: Command) -> bool:
        """
        Check if the user is authorized to perform the command in addition to RBAC.
        A root user is always authorized, any other user must have is_active=True.
        """
        if cmd.user is None:
            return False
        return cmd.user.is_active or enum.Role.ROOT in cmd.user.roles

    def retrieve_user_is_root(self, user: model.User) -> bool:
        return enum.Role.ROOT in user.roles



import abc
from uuid import UUID

import numpy as np

from gen_epix.fastapp import BaseRepository
from gen_epix.fastapp.unit_of_work import BaseUnitOfWork
from gen_epix.seqdb.domain import DOMAIN
from gen_epix.seqdb.domain import model as model  # forces models to be registered now
from gen_epix.seqdb.domain.enum import ServiceType


class BaseSeqRepository(BaseRepository):
    ENTITIES = DOMAIN.get_dag_sorted_entities(
        service_type=ServiceType.SEQ, persistable=True
    )

    @abc.abstractmethod
    def get_distance_matrix_by_seq_ids(
        self,
        uow: BaseUnitOfWork,
        seq_distance_protocol_id: UUID,
        seq_ids: list[UUID],
    ) -> np.ndarray:
        raise NotImplementedError



from sqlalchemy import Engine, select

from gen_epix.fastapp import BaseUnitOfWork, CrudOperation
from gen_epix.fastapp.repositories import SARepository
from gen_epix.fastapp.repositories.sa import SAUnitOfWork
from gen_epix.omopdb.domain import exc, model
from gen_epix.omopdb.domain.repository.organization import BaseOrganizationRepository
from gen_epix.omopdb.repositories import sa_model
from gen_epix.omopdb.repositories.sa_model.base import (
    DB_METADATA_FIELDS,
    GENERATE_SERVICE_METADATA,
    SERVICE_METADATA_FIELDS,
)


class OrganizationSARepository(SARepository, BaseOrganizationRepository):
    def __init__(self, engine: Engine, **kwargs: dict):
        entities = kwargs.pop("entities", BaseOrganizationRepository.ENTITIES)
        super().__init__(
            engine,
            entities=entities,
            service_metadata_fields=SERVICE_METADATA_FIELDS,
            db_metadata_fields=DB_METADATA_FIELDS,
            generate_service_metadata=GENERATE_SERVICE_METADATA,
            **kwargs,
        )

    def is_existing_user_by_key(
        self, uow: BaseUnitOfWork, user_key: str | None
    ) -> bool:
        if user_key is None:
            return False
        assert isinstance(uow, SAUnitOfWork)
        user_row = uow.session.execute(
            select(sa_model.User.id).where(sa_model.User.email == user_key)
        ).all()
        return True if user_row else False

    def retrieve_user_by_key(self, uow: BaseUnitOfWork, user_key: str) -> model.User:
        users: list[model.User] = self.crud(
            uow,
            None,
            model.User,
            None,
            None,
            CrudOperation.READ_ALL,
        )
        for user in users:
            if user.email == user_key.lower():
                return user
        raise exc.NoResultsError()



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


from typing import ClassVar

from gen_epix.seqdb.domain import DOMAIN, model
from gen_epix.seqdb.domain.command.base import CrudCommand


# organization
class IdentifierIssuerCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.IdentifierIssuer


class DataCollectionCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DataCollection


class DataCollectionSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DataCollectionSet


class DataCollectionSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DataCollectionSetMember


class UserCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.User


class UserInvitationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.UserInvitation


# system
class OutageCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Outage


# seq.metadata
class TreeAlgorithmClassCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.TreeAlgorithmClass


class TreeAlgorithmCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.TreeAlgorithm


class LibraryPrepProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.LibraryPrepProtocol


class AssemblyProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.AssemblyProtocol


class LocusDetectionProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.LocusDetectionProtocol


class SnpDetectionProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SnpDetectionProtocol


class AlignmentProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.AlignmentProtocol


class TaxonomyProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.TaxonomyProtocol


class SeqClassificationProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqClassificationProtocol


class PcrProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.PcrProtocol


class AstProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.AstProtocol


class SeqDistanceProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqDistanceProtocol


class SeqCategorySetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqCategorySet


class SeqCategoryCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqCategory


class SubtypingSchemeCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SubtypingScheme


class TaxonCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Taxon


class TaxonSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.TaxonSet


class TaxonSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.TaxonSetMember


class LocusCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Locus


class LocusSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.LocusSet


class LocusSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.LocusSetMember


class RefAlleleCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RefAllele


class TaxonLocusLinkCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.TaxonLocusLink


class RefSeqCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RefSeq


# seq.persistable
class AlleleCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Allele


class AlleleAlignmentCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.AlleleAlignment


class SampleCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Sample


class ReadSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ReadSet


class RawSeqCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RawSeq


class SeqCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Seq


class SeqAlignmentCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqAlignment


class AlleleProfileCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.AlleleProfile


class RefSnpCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RefSnp


class RefSnpCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RefSnp


class RefSnpSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RefSnpSet


class RefSnpSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RefSnpSetMember


class SeqTaxonomyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqTaxonomy


class PcrMeasurementCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.PcrMeasurement


class AstMeasurementCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.AstMeasurement


class AstPredictionCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.AstPrediction


class SeqDistanceCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqDistance


class SeqClassificationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SeqClassification


class SnpProfileCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SnpProfile


class KmerProfileCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.KmerProfile


class KmerDetectionProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.KmerDetectionProtocol


DOMAIN.register_locals(locals())



from __future__ import annotations

import logging
from typing import Any, Type
from uuid import UUID

from cachetools import TTLCache, cached

from gen_epix.casedb.domain import command, enum, exc, model
from gen_epix.casedb.domain.repository import BaseAbacRepository
from gen_epix.casedb.domain.service import BaseAbacService
from gen_epix.casedb.policies.case_abac_policy import CaseAbacPolicy
from gen_epix.casedb.policies.is_organization_admin_policy import (
    IsOrganizationAdminPolicy,
)
from gen_epix.casedb.policies.read_organization_results_only_policy import (
    ReadOrganizationResultsOnlyPolicy,
)
from gen_epix.casedb.policies.read_self_results_only_policy import (
    ReadSelfResultsOnlyPolicy,
)
from gen_epix.casedb.policies.update_user_policy import UpdateUserPolicy
from gen_epix.fastapp import App, CrudOperation, EventTiming
from gen_epix.fastapp.model import Command, CrudCommand, Policy
from gen_epix.filter import (
    BooleanOperator,
    CompositeFilter,
    EqualsBooleanFilter,
    EqualsUuidFilter,
    UuidSetFilter,
)
from util.util import map_paired_elements


class AbacService(BaseAbacService):

    CACHE_INVALIDATION_COMMANDS: tuple[Type[Command], ...] = (
        command.UserAccessCasePolicyCrudCommand,
        command.UserShareCasePolicyCrudCommand,
        command.OrganizationAccessCasePolicyCrudCommand,
        command.OrganizationShareCasePolicyCrudCommand,
    )

    def __init__(
        self,
        app: App,
        repository: BaseAbacRepository,
        logger: logging.Logger | None = None,
        **kwargs: dict,
    ):
        kwargs["service_type"] = kwargs.get(
            "service_type", BaseAbacService.SERVICE_TYPE  # type:ignore[arg-type]
        )
        super().__init__(app, repository=repository, logger=logger, **kwargs)
        self.repository: BaseAbacRepository

    def register_policies(self) -> None:
        f = self.app.register_policy
        policy: Policy
        command_class: Type[Command]
        policy = IsOrganizationAdminPolicy(self)
        for command_class in BaseAbacService.ORGANIZATION_ADMIN_WRITE_COMMANDS:
            f(command_class, policy, EventTiming.BEFORE)
        policy = UpdateUserPolicy(self)
        for command_class in BaseAbacService.UPDATE_USER_COMMANDS:
            f(command_class, policy, EventTiming.BEFORE)
        policy = CaseAbacPolicy(self)
        for command_class in BaseAbacService.CASE_ABAC_COMMANDS:
            f(command_class, policy, EventTiming.DURING)
        policy = ReadOrganizationResultsOnlyPolicy(self)
        for command_class in BaseAbacService.READ_ORGANIZATION_RESULTS_ONLY_COMMANDS:
            f(command_class, policy, EventTiming.AFTER)
        policy = ReadSelfResultsOnlyPolicy(self)
        for command_class in BaseAbacService.READ_SELF_RESULTS_ONLY_COMMANDS:
            f(command_class, policy, EventTiming.AFTER)

    def crud(self, cmd: CrudCommand) -> Any:
        retval = super().crud(cmd)
        # Invalidate cache
        if issubclass(type(cmd), AbacService.CACHE_INVALIDATION_COMMANDS):
            self._get_user_by_id_cached.cache_clear()  # type:ignore[attr-defined]
            self._get_case_abac_cached.cache_clear()  # type:ignore[attr-defined]
        return retval

    def get_organizations_under_admin(self, user: model.User) -> set[UUID]:
        # TODO: inefficient implementation, retrieving first all objs and then filtering.
        # To be improved with e.g. CQS.
        with self.repository.uow() as uow:
            organization_admin_policies: list[
                model.OrganizationAdminPolicy
            ] = self.repository.crud(
                uow,
                user_id=user.id,
                model_class=model.OrganizationAdminPolicy,
                objs=None,
                obj_ids=None,
                operation=CrudOperation.READ_ALL,
            )  # type: ignore
        return set(
            x.organization_id
            for x in organization_admin_policies
            if x.user_id == user.id and x.is_active
        )

    def retrieve_complete_user(
        self, cmd: command.RetrieveCompleteUserCommand
    ) -> model.CompleteUser:
        user = cmd.user
        if user is None:
            raise exc.UnauthorizedAuthError("Command has no user")
        # Get organization
        organization: model.Organization = self.app.handle(  # type: ignore[assignment]
            command.OrganizationCrudCommand(
                user=user,
                obj_ids=user.organization_id,
                operation=CrudOperation.READ_ONE,
            )
        )
        # Get permissions
        permissions = self.app.user_manager.retrieve_user_permissions(user)
        # # Get user case and data collection access
        # case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
        # assert case_abac is not None
        return model.CompleteUser(
            **user.model_dump(exclude={"organization"}),
            organization=organization,
            permissions=permissions,
            # case_abac=case_abac,
        )

    def retrieve_organization_admin_name_emails(
        self,
        cmd: command.RetrieveOrganizationAdminNameEmailsCommand,
    ) -> list[model.UserNameEmail]:
        if not isinstance(cmd.user, model.User):
            raise exc.ServiceException(
                "Command has no or wrong user type: {cmd.user.__class__.__name__}"
            )
        organization_admin_policies: list[model.OrganizationAdminPolicy] = (
            self.app.handle(
                command.OrganizationAdminPolicyCrudCommand(
                    user=cmd.user,
                    operation=CrudOperation.READ_ALL,
                )
            )
        )
        organization_admin_user_ids = {
            x.user_id
            for x in organization_admin_policies
            if x.organization_id == cmd.user.organization_id
        }
        users = self.app.handle(
            command.UserCrudCommand(
                user=cmd.user,
                obj_ids=list(organization_admin_user_ids),
                operation=CrudOperation.READ_SOME,
            )
        )
        return [
            model.UserNameEmail(
                id=x.id,
                name=x.name,
                email=x.email,
            )
            for x in users
        ]

    def temp_update_user_own_organization(
        self,
        cmd: command.UpdateUserOwnOrganizationCommand,
    ) -> model.User:
        """
        Behaviour:
        - Update User.organization
        - Create UserAccessCasePolicies for the user and their new organization, that
          are the union of all UserAccessCasePolicies for that organization. Analogous
          for UserShareCasePolicies.

        - Delete any OrganizationAdminPolicy for the user and their previous
          organization
        - Delete any UserAccessCasePolicy for the user and their previous organization.
          Analogous for UserShareCasePolicy.
        - Create UserAccessCasePolicies for the user and their new organization, one
          for each OrganizationAccessCasePolicy for the new organization combined with
          each case type set and rights from the user's UserAccessCasePolicies from
          the previous organization. Analogous for OrganizationShareCasePolicy and
          UserShareCasePolicy.
        """
        user = cmd.user
        is_new_user = cmd.is_new_user
        tgt_organization_id = cmd.organization_id
        assert user is not None and user.id is not None
        user_id = user.id

        # Special case: new organization is same as current
        if user.organization_id == tgt_organization_id and not is_new_user:
            return user

        with self.repository.uow() as uow:
            # Get all current user access and share case policies
            user_access_case_policies: list[model.UserAccessCasePolicy] = (
                self.app.handle(
                    command.UserAccessCasePolicyCrudCommand(
                        user=user,
                        obj_ids=None,
                        operation=CrudOperation.READ_ALL,
                        query_filter=EqualsUuidFilter(
                            key="user_id",
                            value=user_id,
                        ),
                    )
                )
            )
            user_share_case_policies: list[model.UserShareCasePolicy] = self.app.handle(
                command.UserShareCasePolicyCrudCommand(
                    user=user,
                    obj_ids=None,
                    operation=CrudOperation.READ_ALL,
                    query_filter=EqualsUuidFilter(
                        key="user_id",
                        value=user_id,
                    ),
                )
            )
            # Get all target organization access and share case policies
            organization_access_case_policies: list[
                model.OrganizationAccessCasePolicy
            ] = self.app.handle(
                command.OrganizationAccessCasePolicyCrudCommand(
                    user=user,
                    obj_ids=None,
                    operation=CrudOperation.READ_ALL,
                    query_filter=EqualsUuidFilter(
                        key="organization_id",
                        value=tgt_organization_id,
                    ),
                )
            )
            organization_share_case_policies: list[
                model.OrganizationShareCasePolicy
            ] = self.app.handle(
                command.OrganizationShareCasePolicyCrudCommand(
                    user=user,
                    obj_ids=None,
                    operation=CrudOperation.READ_ALL,
                    query_filter=EqualsUuidFilter(
                        key="organization_id",
                        value=tgt_organization_id,
                    ),
                )
            )
            # Convert target organization policies to new user policies
            new_user_access_case_policies = [
                model.UserAccessCasePolicy(
                    user_id=user_id, **x.model_dump(exclude={"id"})
                )
                for x in organization_access_case_policies
            ]
            new_user_share_case_policies = [
                model.UserShareCasePolicy(
                    user_id=user_id, **x.model_dump(exclude={"id"})
                )
                for x in organization_share_case_policies
            ]
            # Delete the current user policies
            if user_access_case_policies:
                _ = self.app.handle(
                    command.UserAccessCasePolicyCrudCommand(
                        user=user,
                        obj_ids=[
                            x.id for x in user_access_case_policies  # type:ignore[misc]
                        ],
                        operation=CrudOperation.DELETE_SOME,
                    )
                )
            if user_share_case_policies:
                _ = self.app.handle(
                    command.UserShareCasePolicyCrudCommand(
                        user=user,
                        obj_ids=[
                            x.id for x in user_share_case_policies  # type:ignore[misc]
                        ],
                        operation=CrudOperation.DELETE_SOME,
                    )
                )
            # Store the new user policies
            new_user_access_case_policies = self.app.handle(
                command.UserAccessCasePolicyCrudCommand(
                    user=user,
                    objs=new_user_access_case_policies,  # type:ignore[arg-type]
                    operation=CrudOperation.CREATE_SOME,
                )
            )
            new_user_share_case_policies = self.app.handle(
                command.UserShareCasePolicyCrudCommand(
                    user=user,
                    objs=new_user_share_case_policies,  # type:ignore[arg-type]
                    operation=CrudOperation.CREATE_SOME,
                )
            )
            # Change the user organization
            user.organization_id = tgt_organization_id
            user = self.app.handle(
                command.UserCrudCommand(
                    user=user,
                    objs=user,
                    operation=CrudOperation.UPDATE_ONE,
                )
            )

        # Invalidate cache
        self._get_user_by_id_cached.cache_clear()
        self._get_case_abac_cached.cache_clear()

        return user

    @cached(cache=TTLCache(maxsize=1024, ttl=300))
    def _get_user_by_id_cached(self, user_id: UUID) -> model.User:
        user: model.User = self.app.handle(  # type:ignore[assignment]
            command.UserCrudCommand(
                user=None,
                obj_ids=user_id,
                operation=CrudOperation.READ_ONE,
            )
        )
        return user

    def get_case_abac(self, cmd: command.Command) -> model.CaseAbac:
        if cmd.user is None or cmd.user.id is None:
            raise exc.UnauthorizedAuthError("Command has no user")
        user_id = cmd.user.id
        return self._get_case_abac_cached(user_id)

    @cached(cache=TTLCache(maxsize=1024, ttl=300))
    def _get_case_abac_cached(
        self,
        user_id: UUID,
    ) -> model.CaseAbac:
        user = self._get_user_by_id_cached(user_id)
        organization_id = user.organization_id
        # @ABAC: Special case: user has full access, defined as all active private data collection policies and all active organization access and share case policies
        is_full_access = not enum.RoleSet.GE_APP_ADMIN.value.isdisjoint(user.roles)
        if is_full_access:
            return model.CaseAbac(
                is_full_access=True,
                private_data_collection_ids=set(),
                case_type_access_abacs={},
                case_type_share_abacs={},
            )
        # Get filters for all the policies
        organization_filter = CompositeFilter(
            filters=[
                EqualsBooleanFilter(key="is_active", value=True),
                EqualsUuidFilter(key="organization_id", value=organization_id),
            ],
            operator=BooleanOperator.AND,
        )
        user_filter = CompositeFilter(
            filters=[
                EqualsBooleanFilter(key="is_active", value=True),
                EqualsUuidFilter(key="user_id", value=user_id),
            ],
            operator=BooleanOperator.AND,
        )

        # Retrieve all the policies as well as the case type set members and case type col set members
        with self.repository.uow() as uow:
            # Retrieve organization access and share case policies
            organization_access_case_policies: list[
                model.OrganizationAccessCasePolicy
            ] = self.repository.crud(  # type:ignore[assignment]
                uow,
                user_id=user_id,
                model_class=model.OrganizationAccessCasePolicy,
                objs=None,
                obj_ids=None,
                operation=CrudOperation.READ_ALL,
                filter=organization_filter,
            )
            organization_share_case_policies: list[
                model.OrganizationShareCasePolicy
            ] = self.repository.crud(  # type:ignore[assignment]
                uow,
                user_id=user_id,
                model_class=model.OrganizationShareCasePolicy,
                objs=None,
                obj_ids=None,
                operation=CrudOperation.READ_ALL,
                filter=organization_filter,
            )
            # Retrieve user access and share case policies
            user_access_case_policies: list[model.UserAccessCasePolicy] = (
                self.repository.crud(  # type:ignore[assignment]
                    uow,
                    user_id=user_id,
                    model_class=model.UserAccessCasePolicy,
                    objs=None,
                    obj_ids=None,
                    operation=CrudOperation.READ_ALL,
                    filter=user_filter,
                )
            )
            user_share_case_policies: list[model.UserShareCasePolicy] = (
                self.repository.crud(  # type:ignore[assignment]
                    uow,
                    user_id=user_id,
                    model_class=model.UserShareCasePolicy,
                    objs=None,
                    obj_ids=None,
                    operation=CrudOperation.READ_ALL,
                    filter=user_filter,
                )
            )
            # Retrieve case type cols per case type
            case_type_col_map: dict[UUID, set[UUID]] = map_paired_elements(  # type: ignore[assignment]
                [
                    (x.case_type_id, x.id)
                    for x in self.app.handle(
                        command.CaseTypeColCrudCommand(
                            user=user,
                            objs=None,
                            obj_ids=None,
                            operation=CrudOperation.READ_ALL,
                        ),
                    )
                ],
                as_set=True,
            )
            # Retrieve relevant case type set members and case type col set members
            all_case_policies: list[model.BaseCasePolicy] = (
                organization_access_case_policies  # type:ignore[assignment]
                + organization_share_case_policies
                + user_access_case_policies
                + user_share_case_policies  # type:ignore[operator]
            )
            case_type_set_ids = frozenset(x.case_type_set_id for x in all_case_policies)
            case_type_set_member_map: dict[UUID, set[UUID]] = map_paired_elements(  # type: ignore[assignment]
                [
                    (x.case_type_set_id, x.case_type_id)
                    for x in self.app.handle(
                        command.CaseTypeSetMemberCrudCommand(
                            user=user,
                            objs=None,
                            obj_ids=None,
                            operation=CrudOperation.READ_ALL,
                            query_filter=UuidSetFilter(
                                key="case_type_set_id",
                                members=case_type_set_ids,
                            ),
                        ),
                    )
                ],
                as_set=True,
            )
            all_access_case_policies: list[
                model.OrganizationAccessCasePolicy | model.UserAccessCasePolicy
            ] = (organization_access_case_policies + user_access_case_policies)
            case_type_col_set_ids: set[UUID] = set(
                x.read_case_type_col_set_id
                for x in all_access_case_policies
                if x.read_case_type_col_set_id
            ) | set(
                x.write_case_type_col_set_id
                for x in all_access_case_policies
                if x.write_case_type_col_set_id
            )
            case_type_col_set_member_map: dict[UUID, set[UUID]] = map_paired_elements(  # type: ignore[assignment]
                [
                    (x.case_type_col_set_id, x.case_type_col_id)
                    for x in self.app.handle(
                        command.CaseTypeColSetMemberCrudCommand(
                            user=user,
                            objs=None,
                            obj_ids=None,
                            operation=CrudOperation.READ_ALL,
                            query_filter=UuidSetFilter(
                                key="case_type_col_set_id",
                                members=frozenset(case_type_col_set_ids),
                            ),
                        ),
                    )
                ],
                as_set=True,
            )

        # Create and return the case abac object with the intersection of the rights
        return model.CaseAbac(
            is_full_access=is_full_access,
            case_type_access_abacs=AbacService._get_access_intersect(
                organization_access_case_policies,
                user_access_case_policies,
                case_type_set_member_map,
                case_type_col_map,
                case_type_col_set_member_map,
            ),
            case_type_share_abacs=AbacService._get_share_intersect(
                organization_share_case_policies,
                user_share_case_policies,
                case_type_set_member_map,
            ),
        )

    @staticmethod
    def _get_access_dict(
        case_policies: list[
            model.OrganizationAccessCasePolicy | model.UserAccessCasePolicy
        ],
        case_type_set_member_map: dict[UUID, set[UUID]],
        case_type_col_map: dict[UUID, set[UUID]],
        case_type_col_set_member_map: dict[UUID, set[UUID]],
    ) -> dict[UUID, dict[UUID, model.CaseTypeAccessAbac]]:
        def _get_case_type_col_ids(
            case_type_col_set_id: UUID | None,
        ) -> set[UUID]:
            return (
                case_type_col_set_member_map.get(case_type_col_set_id, set())
                if case_type_col_set_id
                else set()
            )

        dict_: dict[UUID, dict[UUID, model.CaseTypeAccessAbac]] = {}
        for x in case_policies:
            case_type_ids = case_type_set_member_map.get(x.case_type_set_id, set())
            is_private = (
                getattr(x, "is_private", True) if hasattr(x, "is_private") else True
            )
            # Create case type access abac object for each case type id
            for case_type_id in case_type_ids:
                if case_type_id not in dict_:
                    dict_[case_type_id] = {}
                all_case_type_col_ids = case_type_col_map[case_type_id]
                case_type_access_abac = model.CaseTypeAccessAbac(
                    case_type_id=case_type_id,
                    data_collection_id=x.data_collection_id,
                    is_private=is_private,
                    add_case=x.add_case,
                    remove_case=x.remove_case,
                    read_case_type_col_ids=all_case_type_col_ids
                    & _get_case_type_col_ids(x.read_case_type_col_set_id),
                    write_case_type_col_ids=all_case_type_col_ids
                    & _get_case_type_col_ids(x.write_case_type_col_set_id),
                    add_case_set=x.add_case_set,
                    remove_case_set=x.remove_case_set,
                    read_case_set=x.read_case_set,
                    write_case_set=x.write_case_set,
                )
                dict_[case_type_id][x.data_collection_id] = case_type_access_abac
        return dict_

    @staticmethod
    def _get_access_intersect(
        organization_access_case_policies: list[model.OrganizationAccessCasePolicy],
        user_access_case_policies: list[model.UserAccessCasePolicy],
        case_type_set_member_map: dict[UUID, set[UUID]],
        case_type_col_map: dict[UUID, set[UUID]],
        case_type_col_set_member_map: dict[UUID, set[UUID]],
    ) -> dict[UUID, dict[UUID, model.CaseTypeAccessAbac]]:
        dict1: dict[UUID, dict[UUID, model.CaseTypeAccessAbac]] = (
            AbacService._get_access_dict(
                organization_access_case_policies,  # type:ignore[arg-type]
                case_type_set_member_map,
                case_type_col_map,
                case_type_col_set_member_map,
            )
        )
        dict2: dict[UUID, dict[UUID, model.CaseTypeAccessAbac]] = (
            AbacService._get_access_dict(
                user_access_case_policies,  # type:ignore[arg-type]
                case_type_set_member_map,
                case_type_col_map,
                case_type_col_set_member_map,
            )
        )
        dict3: dict[UUID, dict[UUID, model.CaseTypeAccessAbac]] = {}

        for case_type_id in set(dict1.keys()) & set(dict2.keys()):
            dict3[case_type_id] = {}
            data1 = dict1[case_type_id]
            data2 = dict2[case_type_id]
            for data_collection_id in set(data1.keys()) & set(data2.keys()):
                # Get both access case policies
                x = data1[data_collection_id]
                y = data2[data_collection_id]
                # Create the case type access abac object with the intersection of
                # the rights
                case_type_access_abac = model.CaseTypeAccessAbac(
                    case_type_id=case_type_id,
                    data_collection_id=data_collection_id,
                    is_private=x.is_private and y.is_private,
                    add_case=x.add_case and y.add_case,
                    remove_case=x.remove_case and y.remove_case,
                    read_case_type_col_ids=x.read_case_type_col_ids
                    & y.read_case_type_col_ids,
                    write_case_type_col_ids=x.write_case_type_col_ids
                    & y.write_case_type_col_ids,
                    add_case_set=x.add_case_set and y.add_case_set,
                    remove_case_set=x.remove_case_set and y.remove_case_set,
                    read_case_set=x.read_case_set and y.read_case_set,
                    write_case_set=x.write_case_set and y.write_case_set,
                )
                # Add to dict only if any rights remain
                if not case_type_access_abac.has_any_rights():
                    continue
                dict3[case_type_id][data_collection_id] = case_type_access_abac
        return dict3

    @staticmethod
    def _get_share_dict(
        case_policies: list[
            model.OrganizationShareCasePolicy | model.UserShareCasePolicy
        ],
        case_type_set_member_map: dict[UUID, set[UUID]],
    ) -> dict[UUID, dict[UUID, model.CaseTypeShareAbac]]:
        dict_: dict[UUID, dict[UUID, model.CaseTypeShareAbac]] = {}
        for x in case_policies:
            # Create case type share abac object for each case type id
            data_collection_id = x.data_collection_id
            from_data_collection_id = x.from_data_collection_id
            case_type_ids = case_type_set_member_map.get(x.case_type_set_id, set())
            for case_type_id in case_type_ids:
                if case_type_id not in dict_:
                    dict_[case_type_id] = {}
                if data_collection_id not in dict_[case_type_id]:
                    dict_[case_type_id][data_collection_id] = model.CaseTypeShareAbac(
                        case_type_id=case_type_id,
                        data_collection_id=data_collection_id,
                        add_case_from_data_collection_ids=set(),
                        remove_case_from_data_collection_ids=set(),
                        add_case_set_from_data_collection_ids=set(),
                        remove_case_set_from_data_collection_ids=set(),
                    )
                case_type_share_abac = dict_[case_type_id][data_collection_id]
                # Add the source data collection id to the appropriate sets
                if x.add_case:
                    case_type_share_abac.add_case_from_data_collection_ids.add(
                        from_data_collection_id
                    )
                if x.remove_case:
                    case_type_share_abac.remove_case_from_data_collection_ids.add(
                        from_data_collection_id
                    )
                if x.add_case_set:
                    case_type_share_abac.add_case_set_from_data_collection_ids.add(
                        from_data_collection_id
                    )
                if x.remove_case_set:
                    case_type_share_abac.remove_case_set_from_data_collection_ids.add(
                        from_data_collection_id
                    )
        return dict_

    @staticmethod
    def _get_share_intersect(
        organization_share_case_policies: list[model.OrganizationShareCasePolicy],
        user_share_case_policies: list[model.UserShareCasePolicy],
        case_type_set_member_map: dict[UUID, set[UUID]],
    ) -> dict[UUID, dict[UUID, model.CaseTypeShareAbac]]:
        dict1: dict[UUID, dict[UUID, model.CaseTypeShareAbac]] = (
            AbacService._get_share_dict(
                organization_share_case_policies,  # type:ignore[arg-type]
                case_type_set_member_map,
            )
        )
        dict2: dict[UUID, dict[UUID, model.CaseTypeShareAbac]] = (
            AbacService._get_share_dict(
                user_share_case_policies,  # type:ignore[arg-type]
                case_type_set_member_map,
            )
        )
        dict3: dict[UUID, dict[UUID, model.CaseTypeShareAbac]] = {}

        for case_type_id in set(dict1.keys()) & set(dict2.keys()):
            dict3[case_type_id] = {}
            data1 = dict1[case_type_id]
            data2 = dict2[case_type_id]
            for data_collection_id in set(data1.keys()) & set(data2.keys()):
                # Get both share case policies
                x = dict1[case_type_id][data_collection_id]
                y = dict2[case_type_id][data_collection_id]
                # Create the case type share abac object with the intersection of the
                # rights
                dict3[case_type_id][data_collection_id] = model.CaseTypeShareAbac(
                    case_type_id=case_type_id,
                    data_collection_id=data_collection_id,
                    add_case_from_data_collection_ids=x.add_case_from_data_collection_ids
                    & y.add_case_from_data_collection_ids,
                    remove_case_from_data_collection_ids=x.remove_case_from_data_collection_ids
                    & y.remove_case_from_data_collection_ids,
                    add_case_set_from_data_collection_ids=x.add_case_set_from_data_collection_ids
                    & y.add_case_set_from_data_collection_ids,
                    remove_case_set_from_data_collection_ids=x.remove_case_set_from_data_collection_ids
                    & y.remove_case_set_from_data_collection_ids,
                )
        # Remove any case type share abac objects with no rights
        for case_type_id in dict3:
            to_pop_data_collection_ids = {
                x for x, y in dict3[case_type_id].items() if not y.has_any_rights()
            }
            for data_collection_id in to_pop_data_collection_ids:
                dict3[case_type_id].pop(data_collection_id)
        return dict3



from gen_epix.casedb.domain import command, enum
from gen_epix.fastapp import Command, PermissionType
from gen_epix.fastapp.services.rbac import BaseRbacService as ServiceBaseRbacService


class BaseRbacService(ServiceBaseRbacService):
    SERVICE_TYPE = enum.ServiceType.RBAC

    NO_RBAC_PERMISSIONS: set[tuple[type[Command], PermissionType]] = {
        # The RegisterInvitedUserCommand is a special case since it is used to create a
        # user and hence no existing user can be included in the command. The command
        # therefore does not require a permission policy, but the permissions can be
        # checked nonetheless through the implementing handler.
        (command.RegisterInvitedUserCommand, PermissionType.EXECUTE),
        (command.GetIdentityProvidersCommand, PermissionType.EXECUTE),
        (command.RetrieveOutagesCommand, PermissionType.EXECUTE),
        (command.UpdateUserOwnOrganizationCommand, PermissionType.EXECUTE),
    }

    def register_handlers(self) -> None:
        f = self.app.register_handler
        for command_class in self.app.domain.get_crud_commands_for_service_type(
            self.service_type
        ):
            f(command_class, self.crud)



# pylint: disable=wildcard-import, unused-import
# because this is a package, and imported as such in other modules
from __future__ import annotations

import datetime
import uuid
from enum import Enum

import ulid


class TimestampFactory(Enum):
    DATETIME_NOW = lambda: datetime.datetime.now()


class IdFactory(Enum):
    UUID4 = uuid.uuid4
    ULID = lambda: ulid.api.new().uuid


class ServiceType(Enum):
    ORGANIZATION = "ORGANIZATION"
    AUTH = "AUTH"
    RBAC = "RBAC"
    OMOP = "OMOP"
    SYSTEM = "SYSTEM"


class RepositoryType(Enum):
    DICT = "DICT"
    SA_SQLITE = "SA_SQLITE"
    SA_SQL = "SA_SQL"


class Role(Enum):
    ROOT = "ROOT"
    ADMIN = "ADMIN"
    REFDATA_ADMIN = "REFDATA_ADMIN"
    DATA_READER = "DATA_READER"
    DATA_WRITER = "DATA_WRITER"
    GUEST = "GUEST"


class RoleSet(Enum):
    ALL = frozenset(
        {
            Role.ROOT,
            Role.ADMIN,
            Role.REFDATA_ADMIN,
            Role.DATA_READER,
            Role.DATA_WRITER,
            Role.GUEST,
        }
    )
    APPLICATION = frozenset({Role.ADMIN})
    REFDATA = frozenset({Role.REFDATA_ADMIN})
    OPERATIONAL = frozenset({Role.DATA_READER, Role.DATA_WRITER, Role.GUEST})


class ConceptSetType(Enum):
    CONTEXT_FREE_GRAMMAR_JSON = "CONTEXT_FREE_GRAMMAR_JSON"
    CONTEXT_FREE_GRAMMAR_XML = "CONTEXT_FREE_GRAMMAR_XML"
    REGULAR_LANGUAGE = "REGULAR_LANGUAGE"
    NOMINAL = "NOMINAL"
    ORDINAL = "ORDINAL"
    INTERVAL = "INTERVAL"



from typing import ClassVar, Self
from uuid import UUID

from pydantic import Field, model_validator

from gen_epix.fastapp import Entity
from gen_epix.seqdb.domain import DOMAIN, enum
from gen_epix.seqdb.domain.model.base import Model
from gen_epix.seqdb.domain.model.seq.metadata import RefSnp
from gen_epix.seqdb.domain.model.seq.persistable import (
    AstMeasurement,
    AstPrediction,
    PcrMeasurement,
    ReadSet,
    SeqClassification,
    SeqDistanceProtocol,
    SeqTaxonomy,
)

_SERVICE_TYPE = enum.ServiceType.SEQ
_ENTITY_KWARGS = {
    "service_type": _SERVICE_TYPE,
    "schema_name": _SERVICE_TYPE.value.lower(),
}


class CompleteContig(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="complete_contigs",
        persistable=False,
        **_ENTITY_KWARGS,
    )
    seq_id: UUID
    seq: str
    qc: enum.QualityControlResult
    index: int


class CompleteAlleleProfile(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="complete_allele_profiles",
        persistable=False,
        **_ENTITY_KWARGS,
    )
    seq_id: UUID
    locus_set_id: UUID
    locus_ids: list[UUID]
    allele_ids: list[UUID | None]
    multiple_allele_ids: dict[UUID, list[UUID]]
    allele_count_by_qc: dict[enum.QualityControlResult, int]


class CompleteSnpProfile(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="complete_snp_profiles",
        persistable=False,
        **_ENTITY_KWARGS,
    )
    seq_id: UUID
    ref_snps: list[RefSnp]
    snps: str
    snp_profile: str
    snp_profile_format: enum.SnpProfileFormat


class CompleteSeq(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="complete_seqs",
        persistable=False,
        **_ENTITY_KWARGS,
    )
    sample_id: UUID | None
    primary_taxon_id: UUID | None
    read_sets: list[ReadSet] | None
    contigs: list[CompleteContig] | None
    allele_profiles: list[CompleteAlleleProfile] | None
    snp_profiles: list[CompleteSnpProfile] | None
    taxa: list[SeqTaxonomy] | None
    ast_predictions: list[AstPrediction] | None
    classifications: list[SeqClassification] | None
    qc: enum.QualityControlResult | None


class CompleteSample(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="complete_samples",
        persistable=False,
        **_ENTITY_KWARGS,
    )
    primary_seq_id: UUID | None
    primary_taxon_id: UUID | None
    seqs: list[CompleteSeq] | None
    pcr_measurements: list[PcrMeasurement] | None
    ast_measurements: list[AstMeasurement] | None


class PhylogeneticTree(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="phylogenetic_trees",
        persistable=False,
        **_ENTITY_KWARGS,
    )
    tree_algorithm: enum.TreeAlgorithm = Field(
        default=None, description="The tree algorithm"
    )
    seq_distance_protocol_id: UUID = Field(
        description="The ID of the sequence distance protocol. FOREIGN KEY"
    )
    seq_distance_protocol: SeqDistanceProtocol = Field(
        default=None, description="The sequence distance protocol"
    )
    leaf_names: list[str] | None = Field(
        default=None,
        description="The list of names of the leaves of the phylogenetic tree to be put in the tree representation instead of seq_ids. Must have the same length as seq_ids.",
    )
    seq_ids: list[UUID] | None = Field(
        default=None,
        description="The list of unique identifiers of the sequence of each leaf of the phylogenetic tree.",
    )
    newick_repr: str = Field(
        description="The Newick representation of the phylogenetic tree."
    )

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        if self.leaf_names:
            if len(set(self.leaf_names)) < len(self.leaf_names):
                raise ValueError("Duplicate leaf_codes")
        if self.seq_ids:
            if len(set(self.seq_ids)) < len(self.seq_ids):
                raise ValueError("Duplicate seq_ids")
            if self.leaf_names and len(self.seq_ids) != len(self.leaf_names):
                raise ValueError(
                    "seq_ids and leaf_codes must have the same length if leaf_codes is provided."
                )
        return self


class MultipleAlignment(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="multiple_alignments",
        persistable=False,
        **_ENTITY_KWARGS,
    )
    alignment_protocol_id: UUID
    seq_ids: list[UUID]
    n_seqs: int
    n_contigs: list[int]
    contig_seqs: list[list[str]]
    n_alignments: int
    n_columns: list[int]
    start_columns: list[list[int]]
    contig_ordinals: list[list[int]]
    contig_start_positions: list[list[int]]
    contig_directions: list[list[bool]]
    lengths: list[list[int]]


DOMAIN.register_locals(locals(), service_type=_SERVICE_TYPE)



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


from typing import ClassVar

import gen_epix.casedb.domain.model.case.case
from gen_epix.casedb.domain import DOMAIN, model
from gen_epix.casedb.domain.command.base import CrudCommand
from gen_epix.fastapp import PermissionTypeSet


# geo
class RegionSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RegionSet


class RegionCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Region


class RegionRelationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RegionRelation


class RegionSetShapeCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RegionSetShape


# ontology
class ConceptCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Concept


class ConceptSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConceptSet


class ConceptSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConceptSetMember


class DiseaseCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Disease


class EtiologicalAgentCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.EtiologicalAgent


class EtiologyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Etiology


# organization
class OrganizationCrudCommand(CrudCommand):
    PERMISSION_TYPE_SET: ClassVar = PermissionTypeSet.CRU
    MODEL_CLASS: ClassVar = model.Organization


class UserCrudCommand(CrudCommand):
    PERMISSION_TYPE_SET: ClassVar = PermissionTypeSet.CRU
    MODEL_CLASS: ClassVar = model.User


class OrganizationSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.OrganizationSet


class OrganizationSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.OrganizationSetMember


class SiteCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Site


class ContactCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Contact


class IdentifierIssuerCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.IdentifierIssuer


class DataCollectionCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DataCollection


class DataCollectionSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DataCollectionSet


class DataCollectionSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DataCollectionSetMember


class DataCollectionRelationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DataCollectionRelation


# rbac
class UserInvitationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.UserInvitation


# subject
class SubjectCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Subject


class SubjectIdentifierCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SubjectIdentifier


# case
class TreeAlgorithmClassCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = gen_epix.casedb.domain.model.case.case.TreeAlgorithmClass


class TreeAlgorithmCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = gen_epix.casedb.domain.model.case.case.TreeAlgorithm


class GeneticDistanceProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = (
        gen_epix.casedb.domain.model.case.case.GeneticDistanceProtocol
    )


class CaseTypeCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseType


class CaseTypeSetCategoryCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseTypeSetCategory


class CaseTypeSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseTypeSet


class CaseTypeSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseTypeSetMember


class DimCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Dim


class ColCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Col


class CaseTypeColSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseTypeColSet


class CaseTypeColSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseTypeColSetMember


class CaseTypeColCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseTypeCol


class CaseCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Case


class CaseDataCollectionLinkCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseDataCollectionLink


class CaseSetCategoryCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseSetCategory


class CaseSetStatusCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseSetStatus


class CaseSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseSet


class CaseSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseSetMember


class CaseSetDataCollectionLinkCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseSetDataCollectionLink


# abac
class OrganizationAdminPolicyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.OrganizationAdminPolicy


class OrganizationAccessCasePolicyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.OrganizationAccessCasePolicy


class UserAccessCasePolicyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.UserAccessCasePolicy


class OrganizationShareCasePolicyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.OrganizationShareCasePolicy


class UserShareCasePolicyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.UserShareCasePolicy


# system
class OutageCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Outage


DOMAIN.register_locals(locals())
